import time
from arcgis._impl.common._mixins import PropertyMap
from arcgis.features import FeatureLayer, FeatureLayerCollection
from arcgis.features._version import Version, VersionManager

########################################################################
class ParcelFabricManager(object):
    """
    The Parcel Fabric Server is responsible for exposing parcel management
    capabilities to support a variety of workflows from different clients
    and systems.

    ====================     ====================================================================
    **Argument**             **Description**
    --------------------     --------------------------------------------------------------------
    url                      Required String. The URI to the service endpoint.
    --------------------     --------------------------------------------------------------------
    gis                      Required GIS. The enterprise connection.
    --------------------     --------------------------------------------------------------------
    version                  Required Version. This is the version object where the modification
                             will occur.
    --------------------     --------------------------------------------------------------------
    flc                      Required FeatureLayerCollection. This is the parent container for
                             ParcelFabricManager.
    ====================     ====================================================================

    """
    _con = None
    _flc = None
    _gis = None
    _url = None
    _version = None
    _properties = None
    #----------------------------------------------------------------------
    def __init__(self,
                 url,
                 gis,
                 version,
                 flc):
        """Constructor"""
        self._url = url
        self._gis = gis
        self._con = gis._portal.con
        self._version = version
        self._flc = flc
    #----------------------------------------------------------------------
    def __str__(self):
        return "<ParcelFabricManager @ %s>" % self._url
    #----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()
    #----------------------------------------------------------------------
    def __enter__(self):
        return self
    #----------------------------------------------------------------------
    def __exit__(self, type, value, traceback):
        return
    #----------------------------------------------------------------------
    @property
    def layer(self):
        """returns the Parcel Layer for the service"""
        if "controllerDatasetLayers" in self._flc.properties and \
           "parcelLayerId" in self._flc.properties.controllerDatasetLayers:
            url = "%s/%s" % (self._flc.url,
                             self._flc.properties.controllerDatasetLayers.parcelLayerId)
            return FeatureLayer(url=url, gis=self._gis)
        return None
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """returns the properties of the service"""
        if self._properties is None:

            res = self._con.get(self._url, {'f':'json'})
            self._properties = PropertyMap(res)
        return self._properties
    #----------------------------------------------------------------------
    def assign_to_record(self,
                         features,
                         record,
                         write_attribute,
                         moment=None):
        """
        Assigns the specified parcel features to the specified record. If
        parcel polygons are assigned, the record polygon will be updated to
        match the cumulative geometry of all the parcels associated to it.
        The Created By Record or Retired By Record attribute field of the
        parcel features is updated with the global ID of the assigned
        record.

        ====================     ====================================================================
        **Argument**             **Description**
        --------------------     --------------------------------------------------------------------
        features                 Required List. The parcel features to assign to the specified record.
                                 Can be parcels, parcel polygons, parcel points, and parcel lines.

                                 Syntax: parcelFeatures=[{"id":"<guid>","layerId":"<layerID>"},{...}]

        --------------------     --------------------------------------------------------------------
        record                   Required String. The record that will be assigned to the specified
                                 parcel features.
        --------------------     --------------------------------------------------------------------
        write_attribute          Required String. Represents the record field to update on the parcel
                                 features. Either the Created By Record or Retired By Record field is
                                 to be updated with the global ID of the assigned record.

                                 Allowed Values: `CreatedByRecord` or `RetiredByRecord`

        --------------------     --------------------------------------------------------------------
        moment                   Optional Integer. This should only be specified by the client when
                                 they do not want to use the current moment
        ====================     ====================================================================

        :returns: Boolean

        """
        url = "{base}/assignFeaturesToRecord".format(base=self._url)
        if moment is None:
            moment = int(time.time())
        params = {
            "gdbVersion" : self._version.properties.versionName,
            "sessionId" : self._version._guid,
            "moment" : moment,
            "parcelFeatures" : features,
            "record" : record,
            "writeAttribute" : write_attribute,
            "f": "json"
        }
        res = self._con.post(url, params)
        if 'success' in res:
            return res['success']
        return res
    #----------------------------------------------------------------------
    def build(self,
              extent=None,
              moment=None,
              return_errors=False,
              record=None):
        """
        A `build` will fix known parcel fabric errors.

        For example, if a parcel polygon exists without lines, then build will
        construct the missing lines. If lines are missing, the polygon row(s)
        are created. When constructing this objects, build will attribute the
        related keys as appropriate. Build also maintains `lineage` and `record`
        features. The parcel fabric must have sufficient information for build
        to work correctly. Ie, source reference document, and connected lines.

        Build provides options to increase performance. The process can just
        work on specific parcels, geometry types or only respond to parcel point
        movement in the case of an adjustment.

        ====================     ====================================================================
        **Argument**             **Description**
        --------------------     --------------------------------------------------------------------
        extent                   Optional Envelope. The extent to build.

                                 Syntax: {"xmin":X min,"ymin": y min, "xmax": x max, "ymax": y max,
                                         "spatialReference": <wkt of spatial reference>}

        --------------------     --------------------------------------------------------------------
        moment                   Optional String. This should only be specified by the client when
                                 they do not want to use the current moment
        --------------------     --------------------------------------------------------------------
        return_errors            Optional Boolean. If True, a verbose response will be given if errors
                                 occured.  The default is False.  **Deprecated**
        --------------------     --------------------------------------------------------------------
        record                   Optional String. Represents the record identifier (guid).  If a
                                 record guid is provided, only parcels associated to the record are
                                 built, regardless of the build extent.
        ====================     ====================================================================


        :return: Boolean

        """
        url = "{base}/build".format(base=self._url)
        if moment is None:
            moment = int(time.time())
        params = {
            "gdbVersion" : self._version.properties.versionName,
            "sessionId" : self._version._guid,
            "moment" : moment,
            "buildExtent" : extent,
            "record" : record,
            "async" : False,
            #"returnErrors" : return_errors,
            "f": "json"
        }
        res =  self._con.post(url, params)
        if 'success' in res:
            return res['success']
        return res
    #----------------------------------------------------------------------
    def clip(self,
             parent_parcels,
             clip_record=None,
             clipping_parcels=None,
             geometry=None,
             moment=None,
             option=None,
             area_unit=None):
        """

        Clip cuts a new child parcel into existing parent parcels. Commonly
        it retires the parent parcel(s) it cuts into to generate a reminder
        child parcel. This type of split is often part of a `parcel split
        metes and bounds` record driven workflow.

        =======================     ====================================================================
        **Argument**                **Description**
        -----------------------     --------------------------------------------------------------------
        parent_parcels              parent parcels that will be clipped into.
                                    Syntax:  parentParcels= <parcel (guid)+layer (name)...>
        -----------------------     --------------------------------------------------------------------
        clip_record                 Optional String. It is the GUID for the active legal record.
        -----------------------     --------------------------------------------------------------------
        clipping_parcels            Optional List. A list of child parcels that will be used to clip
                                    into the parent parcels. Parcel lineage is created if the child
                                    'clipping_parcels' and the parcels being clipped are of the same
                                    parcel type.

                                    Syntax: clippingParcels= < id : parcel guid, layered: <layer id>...>

                                    Example:

                                    [{"id":"{D01D3F47-5FE2-4E39-8C07-E356B46DBC78}","layerId":"16"}]

                                    **Either clipping_parcels or geometry is required.**
        -----------------------     --------------------------------------------------------------------
        geometry                    Optional Polygon. Allows for the clipping a parcel based on geometry instead of
                                    'clippingParcels' geometry. No parcel lineage is created.

                                    **Either clipping_parcels or geometry is required.**
        -----------------------     --------------------------------------------------------------------
        moment                      Optional String. This should only be specified by the client when
                                    they do not want to use the current moment
        -----------------------     --------------------------------------------------------------------
        option                      Optional String. Represents the type of clip to perform:

                                      -  PreserveArea - Preserve the areas that intersect and discard the remainder areas. (default)
                                      -  DiscardArea - Discard the areas that intersect and preserve the remainder areas.
                                      -  PreserveBothAreasSplit - Preserve both the intersecting and remainder areas.
        -----------------------     --------------------------------------------------------------------
        area_unit                   Optional String. Area units to be used when calculating the stated
                                    areas of the clipped parcels. The stated area of the clipped parcels
                                    will be calculated if the stated areas exist on the parent parcels
                                    being clipped.
        =======================     ====================================================================

        :returns: Dictionary


        """
        if moment is None:
            moment = int(time.time())
        gdb_version = self._version.properties.versionName
        session_id = self._version._guid
        url = "{base}/clip".format(base=self._url)
        params = {
            "gdbVersion": gdb_version,
            "sessionId": session_id,
            "parentParcels": parent_parcels,
            "moment" : moment,
            "record" : clip_record,
            "clippingParcels" : clipping_parcels,
            "clippingGeometry" : geometry,
            "clipOption" : option,
            "defaultAreaUnit" : area_unit,
            "f": "json"
        }
        return self._con.post(url, params)
    #----------------------------------------------------------------------
    def merge(self,
              parent_parcels,
              target_parcel_type,
              attribute_overrides=None,
              child_name=None,
              default_area_unit=None,
              merge_record=None,
              merge_into=None,
              moment=None):
        """
        Merge combines 2 or more parent parcels into onenew child parcel. Merge
        sums up legal areas of parent parcels to the new child parcel legal
        area (using default area units as dictated by client). The child parcel
        lines arecomposed from the outer boundaries of the parent parcels.
        Merge can create multipart parcels as well as proportion lines (partial
        overlap of parent parcels). Record footprint is updated to match the
        child parcel.

        ====================     ====================================================================
        **Argument**             **Description**
        --------------------     --------------------------------------------------------------------
        parent_parcels           Required String. It is the parcel(guid)+layer(name) identifiers to
                                 merge.
        --------------------     --------------------------------------------------------------------
        target_parcel_type       Required String. Layer where parcel is merged to.  History is
                                 created when parents and child are of the same parcel type
        --------------------     --------------------------------------------------------------------
        attribute_overrides      Optional List. A list of attributes to set on the child parcel, if
                                 they exist. Pairs of field name and value.

                                 Syntax: attributeOverrides= [{ "type":"PropertySet","propertySetItems":[<field name>,<field value>]}]

                                 * to set subtype, include subtype value in this list.
        --------------------     --------------------------------------------------------------------
        child_name               Optional String. A descript of the child layer. **DEPRECATED**
        --------------------     --------------------------------------------------------------------
        default_area_unit        Optional String. The area units of the child parcel.
        --------------------     --------------------------------------------------------------------
        merge_record             Optional String. Record identifier (guid).  If missing, no history
                                 is created.
        --------------------     --------------------------------------------------------------------
        merge_into               Optional String. A parcel identifier (guid). Invalid to have a
                                 record id.
        --------------------     --------------------------------------------------------------------
        moment                   Optional String. This parameter represents the session moment (the
                                 default is the version current moment). This should only be
                                 specified by the client when they do not want to use the current
                                 moment.
        --------------------     --------------------------------------------------------------------
        area_unit                Optional Integer. Represents the default area units to be used when
                                 calculating the stated area of the merged parcel. The stated area of
                                 the merged parcel will be calculated if the stated areas exist on
                                 the parcels being merged.
        --------------------     --------------------------------------------------------------------
        attribute_overrides      Optional Dict. Represents a list of attributes to set on the new
                                 merged parcel.

                                 Syntax: attribute_overrides={"type":"PropertySet",
                                 "propertySetItems":["<FieldName>",<value>,
                                                    "<FieldName>",<value>,.....,"IsSeed",0]}

        ====================     ====================================================================


        :return: Dictionary

        """
        if moment is None:
            moment = int(time.time())
        gdb_version = self._version.properties.versionName
        session_id = self._version._guid
        url = "{base}/merge".format(base=self._url)
        params = {
            "gdbVersion" : gdb_version, #
            "sessionId" : session_id, #
            "parentParcels" : parent_parcels, #
            "record" : merge_record, #
            "moment" : moment, #
            "targetParcelType" : target_parcel_type,#
            "mergeInto" : merge_into, #
            #"childName" : child_name,
            "defaultAreaUnit" : default_area_unit,#
            "attributeOverrides" : attribute_overrides,#
            "f": "json"
        }
        return self._con.post(url, params)
    #----------------------------------------------------------------------
    def copy_lines_to_parcel_type(self,
                                  parent_parcels,
                                  record,
                                  target_type,
                                  moment=None,
                                  mark_historic=False,
                                  use_source_attributes=False,
                                  attribute_overrides=None,
                                  use_polygon_attributes=False,
                                  parcel_subtype=None):
        """

        Copy lines to parcel type is used when the construction of the
        child parcel is based on parent parcel geometry. It creates a
        copy of the parent parcels lines that the user can modify (insert,
        delete, update) before they build the child parcels. If the source
        parcel type and the target parcel type are identical (common)
        parcel lineage is created.

        =======================     ====================================================================
        **Argument**                **Description**
        -----------------------     --------------------------------------------------------------------
        parent_parcels              Required String. Parcel parcels from which lines are copied.
        -----------------------     --------------------------------------------------------------------
        record                      Required String. The unique identifier (guid) of the active legal
                                    record.
        -----------------------     --------------------------------------------------------------------
        target_type                 Required String. The target parcel layer to which the lines will be
                                    copied to.
        -----------------------     --------------------------------------------------------------------
        moment                      Optional String. This parameter represents the session moment (the
                                    default is the version current moment). This should only be
                                    specified by the client when they do not want to use the current
                                    moment.
        -----------------------     --------------------------------------------------------------------
        mark_historic               Optional Boolean. Mark the parent parcels historic. The default is
                                    False.
        -----------------------     --------------------------------------------------------------------
        use_source_attributes       Optional Boolean. If the source and the target line schema match,
                                    attributes from the parent parcel lines will be copied to the new
                                    child parcel lines when it is set to  True. The default is False.
        -----------------------     --------------------------------------------------------------------
        use_polygon_attributes      Optional Boolean. Parameter representing whether to preserve and
                                    transfer attributes of the parent parcels to the generated seeds.
        -----------------------     --------------------------------------------------------------------
        attribute_overrides         Optional Dictionary. To set fields on the child parcel lines with a
                                    specific value. Uses a key/value pair of FieldName/Value.

                                    Example:

                                    {'type' : "PropertySet", "propertySetItems" : []}
        -----------------------     --------------------------------------------------------------------
        parcel_subtype              Optional Integer. Represents the target parcel subtype.
        =======================     ====================================================================

        :returns: boolean


        """
        if moment is None:
            moment = int(time.time())
        gdb_version = self._version.properties.versionName
        session_id = self._version._guid
        url = "{base}/copyLinesToParcelType".format(base=self._url)
        params = {
            "gdbVersion": gdb_version,
            "sessionId": session_id,
            "parentParcels": parent_parcels,
            "record" : record,
            "markParentAsHistoric" : mark_historic,
            "useSourceLineAttributes": use_source_attributes,
            "useSourcePolygonAttributes" : use_polygon_attributes,
            "targetParcelType" : target_type,
            "targetParcelSubtype" : parcel_subtype,
            "attributeOverrides": attribute_overrides,
            "moment" : moment,
            "f": "json"
        }
        return self._con.post(url, params)
    #----------------------------------------------------------------------
    def change_type(self,
                    parcels,
                    target_type,
                    parcel_subtype=0,
                    moment=None):
        """

        Changes a set of parcels to a new parcel type. It creates new
        polygons and lines and deletes them from the source type. This
        is used when a parcel was associated in the wrong parcel type subtype
        and/or when creating multiple parcels as part of a build process.
        Example: when lot parcels are created as part of a subdivision, the
        road parcel is moved to the encumbrance (easement) parcel type.

        =======================     ====================================================================
        **Argument**                **Description**
        -----------------------     --------------------------------------------------------------------
        parcels                     Required List. Parcels list that will change type
        -----------------------     --------------------------------------------------------------------
        target_type                 Required String. The target parcel layer
        -----------------------     --------------------------------------------------------------------
        target_subtype              Optional Integer. Target parcel subtype. The default is 0 meaning
                                    no subtype required.
        -----------------------     --------------------------------------------------------------------
        moment                      Optional String. This parameter represents the session moment (the
                                    default is the version current moment). This should only be
                                    specified by the client when they do not want to use the current
                                    moment.
        =======================     ====================================================================

        :returns: Dictionary


        """
        if moment is None:
            moment = int(time.time())
        gdb_version = self._version.properties.versionName
        session_id = self._version._guid
        url = "{base}/changeParcelType".format(base=self._url)
        params = {
            "gdbVersion": gdb_version,
            "sessionId": session_id,
            "parcels" : parcels,
            "targetParcelType" : target_type,
            "targetParcelSubtype" : parcel_subtype,
            "moment" : moment,
            "f": "json"
        }
        res = self._con.post(url, params)
        if 'success' in res:
            return res['success']
        return res
    #----------------------------------------------------------------------
    def delete(self, parcels, moment=None):
        """

        Delete a set of parcels, removing associated or unused lines, and
        connected points.

        =======================     ====================================================================
        **Argument**                **Description**
        -----------------------     --------------------------------------------------------------------
        parcels                     Required List. The parcels to erase.
        -----------------------     --------------------------------------------------------------------
        moment                      Optional String. This parameter represents the session moment (the
                                    default is the version current moment). This should only be
                                    specified by the client when they do not want to use the current
                                    moment.
        =======================     ====================================================================

        :returns: Boolean


        """
        if moment is None:
            moment = int(time.time())
        gdb_version = self._version.properties.versionName
        session_id = self._version._guid
        url = "{base}/deleteParcels".format(base=self._url)
        params = {
            "gdbVersion": gdb_version,
            "sessionId": session_id,
            "parcels" : parcels,
            "moment" : moment,
            "f": "json"
        }
        return self._con.post(url, params)['success']
    #----------------------------------------------------------------------
    def update_history(self, features, record,
                       moment=None, set_as_historic=False):
        """
        Sets the specified parcel features to current or historic using the
        specified record. If setting current parcels as historic, the
        Retired By Record field of the features is updated with the Global
        ID of the specified record. If setting historic parcels as current,
        the Created By Record field of the features is updated with the
        Global ID of the specified record.

        =======================     ====================================================================
        **Argument**                **Description**
        -----------------------     --------------------------------------------------------------------
        features                    Required List. The parcel features to be set as historic or current.
                                    Can be parcels, parcel polygons, parcel points, and parcel lines.

                                    Syntax: ```features=[{"id":"<guid>","layerId":"<layerID>"},{...}]```
        -----------------------     --------------------------------------------------------------------
        record                      Required String. A **GUID** representing the record that will be
                                    assigned to the features set as current or historic.
        -----------------------     --------------------------------------------------------------------
        moment                      Optional String. This parameter represents the session moment (the
                                    default is the version current moment). This should only be
                                    specified by the client when they do not want to use the current
                                    moment.
        -----------------------     --------------------------------------------------------------------
        set_as_historic             Optional Boolean.  Boolean parameter representing whether to set the
                                    features as historic (true). If false, features will be set as
                                    current.
        =======================     ====================================================================

        :returns: Dictionary

        """
        if moment is None:
            moment = int(time.time())
        gdb_version = self._version.properties.versionName
        session_id = self._version._guid
        url = "{base}/updateParcelHistory".format(base=self._url)
        params = {
            "gdbVersion": gdb_version,
            "sessionId": session_id,
            "moment" : moment,
            'record' : record,
            'setAsHistoric' : set_as_historic,
            'parcelFeatures' : features,
            "f": "json"
        }
        return self._con.post(url, params)
    #----------------------------------------------------------------------
    def create_seeds(self,
                     record,
                     moment=None,
                     extent=None):
        """

        Create seeds creates parcel seeds for closed loops of lines that
        are associated with the specified record.

        When building parcels from lines, parcel seeds are used. A parcel
        seed is the initial state or seed state of a parcel. A parcel seed
        indicates to the build process that a parcel can be built from the
        lines enclosing the seed.

        A parcel seed is a minimized polygon feature and is stored in the
        parcel type polygon feature class.

        =======================     ====================================================================
        **Argument**                **Description**
        -----------------------     --------------------------------------------------------------------
        record                      Required String. A **GUID** representing the record that will be
                                    assigned to the features set as current or historic.
        -----------------------     --------------------------------------------------------------------
        moment                      Optional String. This parameter represents the session moment (the
                                    default is the version current moment). This should only be
                                    specified by the client when they do not want to use the current
                                    moment.
        -----------------------     --------------------------------------------------------------------
        extent                      Optional Dict/arcgis.Geometry.Envelope. The envelope of the extent
                                    in which to create seeds.
        =======================     ====================================================================

        :returns: Dictionary

        """
        from arcgis.geometry import Envelope
        if isinstance(extent, (dict, Envelope)):
            extent = dict(extent)
        elif extent is None:
            pass
        elif not extent is None:
            raise ValueError("Parameter `extent` must be None, Envelope or dict.")
        if moment is None:
            moment = int(time.time())
        gdb_version = self._version.properties.versionName
        session_id = self._version._guid
        url = "{base}/createSeeds".format(base=self._url)
        params = {
            "gdbVersion": gdb_version,
            "sessionId": session_id,
            "moment" : moment,
            'record' : record,
            'extent' : extent,
            "f": "json"
        }
        return self._con.post(url, params)
    #----------------------------------------------------------------------
    def duplicate(self,
                  parcels,
                  parcel_type,
                  record,
                  parcel_subtype=None,
                  moment=None):
        """
        `duplicate` allows for the cloning of parcels from a specific record.

        Parcels can be duplicated in the following ways:

          -  Duplicate to a different parcel type.
          -  Duplicate to a different subtype in the same parcel type.
          -  Duplicate to a different subtype in a different parcel type.

        Similarly, parcel seeds can be duplicated to subtypes and different parcel types.

        =======================     ====================================================================
        **Argument**                **Description**
        -----------------------     --------------------------------------------------------------------
        parcels                     Required List. A list of parcels to duplicate.

                                    :Syntax:

                                    ```python
                                    [{"id":"<parcelguid>","layerId":"16"},{...}]
                                    ```

        -----------------------     --------------------------------------------------------------------
        parcel_type                 Required Integer. The target parcel type.
        -----------------------     --------------------------------------------------------------------
        record                      Required String. A **GUID** representing the record that will be
                                    assigned to the features set as current or historic.
        -----------------------     --------------------------------------------------------------------
        parcel_subtype              Optional Integer. The target parcel subtype.  The default is 0.
        -----------------------     --------------------------------------------------------------------
        moment                      Optional String. This parameter represents the session moment (the
                                    default is the version current moment). This should only be
                                    specified by the client when they do not want to use the current
                                    moment.
        =======================     ====================================================================

        :returns: Dictionary

        """
        if parcel_type is None:
            parcel_subtype = 0
        if moment is None:
            moment = int(time.time())
        gdb_version = self._version.properties.versionName
        session_id = self._version._guid
        url = "{base}/duplicateParcels".format(base=self._url)
        params = {
            "gdbVersion": gdb_version,
            "sessionId": session_id,
            "moment" : moment,
            'record' : record,
            'moment' : moment,
            'parcels' : parcels,
            'targetParcelType' : parcel_type,
            'targetParcelSubtype' : parcel_subtype,
            "f": "json"
        }
        return self._con.post(url, params)
    # ----------------------------------------------------------------------
    def analyze_least_squares_adjustment(self,
                                    analysis_type="CONSISTENCY_CHECK",
                                    convergence_tolerance=0.05,
                                    parcel_features=None,
                                    future=False):
        """
        Note: Least Squares Adjustment functionality introduced at version 10.8.1

        Analyzes the parcel fabric measurement network by running a least squares adjustment on the
        input parcels. A least-squares adjustment is a mathematical procedure that uses statistical
        analysis to estimate the most likely coordinates for connected points in a measurement network.

        Use apply_least_squares_adjustment to apply the results of a least squares adjustment to parcel fabric feature classes.
        ====================    ====================================================================
        **Argument**            **Description**
        --------------------    --------------------------------------------------------------------
        analysis_type           Optional string. Represents the type of least squares analysis that will be run on the input parcels.

                                    CONSISTENCY_CHECK - A free-network least-squares adjustment will be run to check dimensions on
                                    parcel lines for inconsistencies and mistakes. Fixed or weighted control points will not be
                                    used by the adjustment.

                                    WEIGHTED_LEAST_SQUARES - A weighted least-squares adjustment will be run to compute updated
                                    coordinates for parcel points. The parcels being adjusted should connect to at least two fixed
                                    or weighted control points.

                                The default value is CONSISTENCY_CHECK.
        --------------------    --------------------------------------------------------------------
        convergence_tolerance   Optional float. Represents the maximum coordinate shift expected after iterating the least squares adjustment. A least
                                squares adjustment is run repeatedly (in iterations) until the solution converges. The solution is
                                considered converged when maximum coordinate shift encountered becomes less than the specified convergence
                                tolerance.

                                The default value is 0.05 meters or 0.164 feet.
        --------------------    --------------------------------------------------------------------
        parcel_features         Optional list. Represents the input parcels that will be analyzed by a least squares adjustment.

                                    Syntax: parcel_features = [{"id":"<guid>","layerId":"<layerID>"},{...}]

                                If None, the method will analyze the entire parcel fabric.
        --------------------    --------------------------------------------------------------------
        future                  Optional boolean. If true, the request is processed as an asynchronous job and a URL is returned that points a location
                                displaying the status of the job.

                                The default is False.
        ====================    ====================================================================
        :return: Dictionary

        """
        url = "{base}/analyzeByLeastSquaresAdjustment".format(base=self._url)
        params = {
            "gdbVersion": self._version.properties.versionName,
            "sessionId": self._version._guid,
            "analysisType": analysis_type,
            "convergenceTolerance": convergence_tolerance,
            "parcelFeatures": parcel_features,
            "async": future,
            "f": "json"
            }
        if future:
            res = self._con.post(path=url, postdata=params)
            f = self._run_async(self._status_via_url, con=self._con,
                                url=res["statusUrl"], params={"f": "json"})
            return f
        else:
            res = self._con.post(url, params)
            if 'success' in res:
                return res
            return res
    # ----------------------------------------------------------------------
    def apply_least_squares_adjustment(self,
              movement_tolerance=0.05,
              update_attributes=True,
              future=False):
        """
        Note: Least Squares Adjustment functionality introduced at version 10.8.1

        Applies the results of a least squares adjustment to parcel fabric feature classes. Least squares adjustment results stored
        in the AdjustmentLines and AdjustmentPoints feature classes are applied to the corresponding parcel line, connection line,
        and parcel fabric point feature classes.

        Use analyze_least_squares_adjustment to run a least-squares analysis on parcels and store the results in adjustment feature classes.
        ====================     ====================================================================
        **Argument**             **Description**
        --------------------     --------------------------------------------------------------------
        movement_tolerance        Optional float. Represents the minimum allowable coordinate shift when updating parcel fabric points. If the distance between
                                  the adjustment point and the parcel fabric point is greater than the specified tolerance, the parcel fabric
                                  point is updated to the location of the adjustment point.

                                  The default tolerance is 0.05 meters or 0.164 feet.
        --------------------     --------------------------------------------------------------------
        update_attributes         Optional boolean. Specifies whether attribute fields in the parcel fabric Points feature class will be updated with
                                  statistical metadata. The XY Uncertainty, Error Ellipse Semi Major, Error Ellipse Semi Minor, and
                                  Error Ellipse Direction fields will be updated with the values stored in the same fields in the AdjustmentPoints
                                  feature class.

                                  The default is True
        --------------------     --------------------------------------------------------------------
        future                   Optional boolean. If true, the request is processed as an asynchronous job and a URL is returned that points a location
                                 displaying the status of the job.

                                 The default is False.
        ====================     ====================================================================

        :return: Dictionary

        """

        url = "{base}/applyLeastSquaresAdjustment".format(base=self._url)
        params = {
            "gdbVersion": self._version.properties.versionName,
            "sessionId": self._version._guid,
            "movementTolerance": movement_tolerance,
            "updateAttributes": update_attributes,
            "async": future,
            "f": "json"
            }
        if future:
            res = self._con.post(path=url, postdata=params)
            future = self._run_async(self._status_via_url, con=self._con,
                                url=res["statusUrl"], params={"f": "json"})
            return future
        else:
            res = self._con.post(url, params)
            if 'success' in res:
                return res
            return res
    # ----------------------------------------------------------------------
    def _run_async(self, fn, **inputs):
        """runs the inputs asynchronously"""
        import concurrent.futures
        tp = concurrent.futures.ThreadPoolExecutor(1)
        future = tp.submit(fn=fn, **inputs)
        tp.shutdown(False)
        return future
    # ----------------------------------------------------------------------
    def _status_via_url(self, con, url, params):
        """
        performs the asynchronous check to see if the operation finishes
        """
        status_allowed = ['esriJobSubmitted', 'esriJobWaiting', 'esriJobExecuting', 'esriJobSucceeded',
                          'esriJobFailed', 'esriJobTimedOut', 'esriJobCancelling', 'esriJobCancelled']
        status = con.get(url, params)
        while status['status'] in status_allowed and status["status"] != "esriJobSucceeded":
            if status['status'] == 'esriJobSucceeded':
                return status
            elif status['status'] in ['esriJobFailed', 'esriJobTimedOut', 'esriJobCancelled']:
                break
            status = con.get(url, params)
        return status


"""
This file can contain various metrics for different types of deep learning models. 
"""

# MultiLabel Image Classification

def flatten_check(inp, targ):
    "Check that `out` and `targ` have the same number of elements and flatten them."
    inp,targ = inp.contiguous().view(-1),targ.contiguous().view(-1)
    assert(len(inp) == len(targ))
    return inp,targ

def accuracy_multi(inp, targ, thresh=0.5, sigmoid=True):
    "Compute accuracy when `inp` and `targ` are the same size."
    inp,targ = flatten_check(inp,targ)
    if sigmoid: inp = inp.sigmoid()
    return ((inp>thresh)==targ.bool()).float().mean()


import contextlib
import OpenSSL.crypto
import os
import ssl
import tempfile

#----------------------------------------------------------------------
def pfx_to_pem(pfx_path, pfx_password):
    """ Decrypts the .pfx file to be used with requests.

    ===============     ====================================================================
    **Argument**        **Description**
    ---------------     --------------------------------------------------------------------
    pfx_path            Required string.  File pathname to .pfx file to parse.
    ---------------     --------------------------------------------------------------------
    pfx_password        Required string.  Password to open .pfx file to extract key/cert.
    ===============     ====================================================================

    :return:
       File path to key_file located in a tempfile location
       File path to cert_file located in a tempfile location
    """
    try:
        import OpenSSL.crypto
    except:
        raise RuntimeError("OpenSSL.crypto library is not installed.  You must install this in order "+\
                           "to use a PFX for connecting to a PKI protected portal.")
    key_file = tempfile.NamedTemporaryFile(suffix='.pem', delete=False)
    cert_file = tempfile.NamedTemporaryFile(suffix='.pem', delete=False)
    k = open(key_file.name, 'wb')
    c = open(cert_file.name, 'wb')
    try:
        pfx = open(pfx_path, 'rb').read()
        p12 = OpenSSL.crypto.load_pkcs12(pfx, pfx_password)
    except OpenSSL.crypto.Error:
        raise RuntimeError("Invalid PFX password.  Unable to parse file.")
    k.write(OpenSSL.crypto.dump_privatekey(OpenSSL.crypto.FILETYPE_PEM, p12.get_privatekey()))
    c.write(OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_PEM, p12.get_certificate()))
    k.close()
    c.close()
    return key_file.name, cert_file.name


#if __name__ == "__main__":
    #cert_file=r"./pki_cert_test.pfx"
    #password="1234"
    #print(pfx_to_pem(cert_file, password))


import os
import sys
import json
import arcgis
from arcgis.learn import FeatureClassifier
import arcpy
prf_root_dir = os.path.join(os.path.dirname(__file__), os.pardir)
sys.path.append(prf_root_dir)
import numpy as np
from .util import normalize_batch

try:
    from fastai.vision import *
    import torch
    HAS_PYTORCH_FA = True

except Exception as e:
    HAS_PYTORCH_FA = False

imagenet_stats = ([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])

imagenet_mean = 255* np.array(imagenet_stats[0], dtype=np.float32)
imagenet_std  = 255* np.array(imagenet_stats[1], dtype=np.float32)

def norm(x, mean=imagenet_mean, std=imagenet_std):
    return (x - mean)/std

def denorm(x, mean=imagenet_mean, std=imagenet_std):
    return x * std + mean

class ChildObjectDetector:

    def initialize(self, model, model_as_file):
        if not HAS_PYTORCH_FA:
            raise Exception('PyTorch (version 1.1.0 or above) and fast.ai (version 1.0.54 or above) libraries are not installed. Install PyTorch using "conda install pytorch=1.1.0 fastai=1.0.54".')

        if model_as_file:
            with open(model, 'r') as f:
                self.emd = json.load(f)
        else:
            self.emd = json.loads(model)

        if arcpy.env.processorType == "GPU" and torch.cuda.is_available():
            self.device = torch.device('cuda')
            arcgis.env._processorType = "GPU"
        else:
            self.device = torch.device('cpu')
            arcgis.env._processorType = "CPU"        

        # Using arcgis.learn FeatureClassifer from_model function.
        self.cf = FeatureClassifier.from_model(emd_path=model)
        self.model = self.cf.learn.model
        self.model.eval()

    def getParameterInfo(self, required_parameters):
        if "MetaDataMode" in self.emd and self.emd["MetaDataMode"] == "MultiLabeled_Tiles":
            required_parameters.append(
                {
                    'name': 'score_threshold',
                    'dataType': 'numeric',
                    'value': 0.5,
                    'required': False,
                    'displayName': 'Confidence Score Threshold [0.0, 1.0]',
                    'description': 'Confidence score threshold value [0.0, 1.0]'
                }
            )
        return required_parameters

    def getConfiguration(self, **scalars):

        if 'BatchSize' not in self.emd and 'batch_size' not in scalars:
            self.batch_size = 1
        elif 'BatchSize' not in self.emd and 'batch_size' in scalars:
            self.batch_size = int(scalars['batch_size'])
        else:
            self.batch_size = int(self.emd['BatchSize'])

        self.thresh = float(scalars.get('score_threshold', 0.5))   # Default 0.5 threshold

        return {
            # CropSizeFixed is a boolean value parameter (1 or 0) in the emd file, representing whether the size of
            # tile cropped around the feature is fixed or not.
            # 1 -- fixed tile size, crop fixed size tiles centered on the feature. The tile can be bigger or smaller
            # than the feature;
            # 0 -- Variable tile size, crop out the feature using the smallest fitting rectangle. This results in tiles
            # of varying size, both in x and y. the ImageWidth and ImageHeight in the emd file are still passed and used
            # as a maximum size. If the feature is bigger than the defined ImageWidth/ImageHeight, the tiles are cropped
            # the same way as in the fixed tile size option using the maximum size.
            'CropSizeFixed': int(self.emd['CropSizeFixed']),

            # BlackenAroundFeature is a boolean value paramater (1 or 0) in the emd file, representing whether blacken
            # the pixels outside the feature in each image tile.
            # 1 -- Blacken
            # 0 -- Not blacken
            'BlackenAroundFeature': int(self.emd['BlackenAroundFeature']),

            'extractBands': tuple(self.emd['ExtractBands']),
            'tx': self.emd['ImageWidth'],
            'ty': self.emd['ImageHeight'],
            'batch_size': self.batch_size
        }

    def vectorize(self, **pixelBlocks):
        
        # Get pixel blocks - tuple of 3-d rasters: ([bands,height,width],[bands,height.width],...)
        # Convert tuple to 4-d numpy array
        batch_images = np.asarray(pixelBlocks['rasters_pixels'])

        # Get the shape of the 4-d numpy array
        batch, bands, height, width = batch_images.shape

        rings = []
        labels, confidences = [], []

        # Normalize Image
        if "NormalizationStats" in self.emd:
            batch_images = normalize_batch(batch_images, self.emd)
        else:
            # Transpose the image dimensions to [batch, height, width, bands],
            # normalize and transpose back to [batch, bands, height, width]
            batch_images = norm(batch_images.transpose(0,2,3,1)).transpose(0, 3, 1, 2)

        # Convert to torch tensor, set device and convert to float
        batch_images = torch.tensor(batch_images).to(self.device).float()

        # the second element in the passed tuple is hardcoded to make fastai's pred_batch work
        predictions = self.cf.learn.pred_batch(batch=(batch_images, torch.tensor([40]).to(self.device)))
        # predictions: torch.tensor(B,C), where B is the batch size and C is the number of classes

        # Using emd to map the class
        class_map = [c['Name'] for c in self.emd["Classes"]]

        # For Multi Label Classification
        if "MetaDataMode" in self.emd and self.emd["MetaDataMode"] == "MultiLabeled_Tiles":
            for pred in predictions:

                # Select the class labels >= threshold and convert them to a comma separated string
                class_idxs = np.where(pred >= self.thresh)[0]
                lbls = [class_map[idx] for idx in class_idxs]
                lbls_string = ";".join(lbls)
                labels.append(lbls_string)

                # Select all confidences and convert them to a comma separated string
                scores = [str(p.item()) for p in pred]
                # scores = [str(pred[idx].tolist()) for idx in class_idxs]
                scores_string = ";".join(scores)
                confidences.append(scores_string)

        # For Single Label Classification
        else:
            # torch.max returns the max value and the index of the max as a tuple
            confidences, class_idxs = torch.max(predictions, dim=1)
            confidences = confidences.tolist()
            labels = [class_map[c] for c in class_idxs]

        # Appending this ring for all the features in the batch
        rings = [[[0, 0], [width - 1, 0], [width - 1, height - 1], [0, height - 1]] for i in range(batch)]

        return rings, confidences, labels



import uuid
import arcgis
from .._utils._basewidget import _BaseWidget
from .._utils._basewidget import NoDataProperties


class Gauge(_BaseWidget):
    """
    Creates a dashboard Gauge widget.

    =========================   ===========================================
    **Argument**                **Description**
    -------------------------   -------------------------------------------
    item                        Required Portal Item object. Item object can
                                be a Feature Layer or a MapWidget.
    -------------------------   -------------------------------------------
    name                        Optional string. Name of the gauge widget.
    -------------------------   -------------------------------------------
    layer                       Optional integer. Layer number when item is
                                a mapwidget.
    -------------------------   -------------------------------------------
    title                       Optional string. Title or Caption for the
                                widget.
    -------------------------   -------------------------------------------
    description                 Optional string. Description for the widget.
    =========================   ===========================================
    """
    def __init__(self, item, name='Gauge', layer=0, title='', description=''):
        super().__init__(name, title, description)
        # General Block

        if item.type not in ['Feature Service', 'mapWidget']:
            raise Exception("Please specify an item")

        self.item = item
        self.type = "gaugeWidget"
        self.layer = layer

        self._max_features = 50
        self._show_last_update = True

        self._nodata = NoDataProperties._nodata_init()
        self._novalue = NoDataProperties._nodata_init()

        self._data = GaugeData._create_data(data_item = item)
        self._mindata = GaugeData._create_data(data_item=item, value_type="fixedvalue")
        self._maxdata = GaugeData._create_data(data_item=item, value_type="fixedvalue")
        self ._gauge = GaugeProperties._gauge_init()
    
    @classmethod
    def _from_json(cls, widget_json):
        gis = arcgis.env.active_gis
        itemid = widget_json["datasets"]["datasource"]["itemid"]
        name = widget_json["name"]
        item = gis.content.get(itemid)
        title = widget_json["caption"]
        description = widget_json["description"]
        gauge = Gauge(item, name, title, description)

        return gauge

    @property
    def data(self):
        """
        :return: Gauge Data object. Set data properties, categories and values.
        """
        return self._data

    @property
    def gauge_options(self):
        """
        :return: Gauge options object. Set gauge properties.
        """
        return self._gauge
    
    @property
    def no_data(self):
        """
        :return: Nodata Object, set various nodata properties
        """
        return self._nodata

    @property
    def max_features(self):
        """
        :return: max features for widget.
        """
        return self._max_features

    @max_features.setter
    def max_features(self, value):
        """
        Set max features for widget.
        """
        self._max_features = int(value)

    @property
    def show_last_update(self):
        """
        :return: show last update or not.
        """
        return self._show_last_update
    
    @show_last_update.setter
    def show_last_update(self, value):
        """
        Set show last update or not.
        """
        self._show_last_update = value
        if not isinstance(value, bool):
            self._show_last_update = True

    def _convert_to_json(self):

        self._arrows = []
        self._bands = []
        self._axes = []
        self._datasets = []
        self._data_statistic_defintion = []
        self._mindata_statistic_defintion = []
        self._mazdata_statistic_defintion = []

        if self._gauge.style == "meter":
            self._arrows.append({
                    "id":"value",
                    "color":None,
                    "axis":"ticks",
                    "alpha":1,
                    "borderAlpha":1,
                    "radius":"100%",
                    "innerRadius":0,
                    "nailRadius":0,
                    "startWidth":8
                })
        if self._gauge.style == "progress":
            self._bands.append({
                            "alpha":1,
                            "color":"#bee8ff",
                            "startValue":0,
                            "endValue":1,
                            "radius":"100%",
                            "innerRadius":"75%",
                            "colorThresholds":[]
                        })
        
        if self._gauge.shape == "horseshoe":
            self._gauge_start_angle = -90
            self._gauge_end_angle = 90
        elif self._gauge.shape == "halfdonut":
            self._gauge_start_angle = -120
            self._gauge_end_angle = 120
        else:
            self._gauge_start_angle = 0
            self._gauge_end_angle = 360       

        self._axes.append({
                    "id":"main",
                    "style":self._gauge.style,
                    "startAngle":self._gauge_start_angle,
                    "endAngle":self._gauge_end_angle,
                    "startValue":0,
                    "endValue":1,
                    "labelsEnabled":True,
                    "labelOffset":0,
                    "color":"#8400a8",
                    "inside":True,
                    "gridInside":True,
                    "axisAlpha":0,
                    "axisColor":"#8400a8",
                    "axisThickness":0,
                    "tickAlpha":0,
                    "tickColor":"#8400a8",
                    "tickLength":0,
                    "tickThickness":0,
                    "minorTickLength":0,
                    "radius":"80%",
                    "bottomText":"",
                    "bands":self._bands
                })
        if self._gauge.style == "meter":
            self._axes.append({
                    "id":"ticks",
                    "style":self._gauge.style,
                    "startAngle":self._gauge_start_angle,
                    "endAngle":self._gauge_end_angle,
                    "startValue":0,
                    "endValue":1,
                    "labelsEnabled":False,
                    "labelOffset":0,
                    "color":"#8400a8",
                    "inside":True,
                    "gridInside":True,
                    "axisAlpha":0.7,
                    "axisColor":"#005ce6",
                    "axisThickness":3,
                    "tickAlpha":0.7,
                    "tickColor":"#005ce6",
                    "tickLength":-12,
                    "tickThickness":3,
                    "minorTickLength":-10,
                    "radius":"80%",
                    "bottomText":"",
                    "bands":self._bands
                })
            self._axes.append(                {
                    "id":"labels",
                    "style":self._gauge.style,
                    "startAngle":self._gauge_start_angle,
                    "endAngle":self._gauge_end_angle,
                    "startValue":0,
                    "endValue":1,
                    "labelsEnabled":True,
                    "labelOffset":32,
                    "fontSize":20,
                    "color":"#ff5500",
                    "inside":False,
                    "gridInside":True,
                    "axisAlpha":0,
                    "axisColor":"#8400a8",
                    "axisThickness":0,
                    "tickAlpha":0,
                    "tickColor":"#8400a8",
                    "tickLength":0,
                    "tickThickness":0,
                    "minorTickLength":0,
                    "radius":"80%",
                    "bottomText":"",
                    "bands":self._bands
                })
        if self._data.value_type == 'statistic':
            self._data_statistic_defintion.append({
                "onStatisticField":self._data.statistics_field,
                "outStatisticFieldName":"value",
                "statisticType":self._data.statistic
            })

        if self.item.type == 'mapWidget':
            wlayer = self.item.layers[self.layer]
            widget_id = self.item._id
            layer_id = wlayer["id"]
            self._datasource = {"id":str(widget_id)+'#'+str(layer_id)}
        else:
            self._datasource = {
                        "type": "featureServiceDataSource",
                        "itemId": self.item.itemid,
                        "layerId": 0,
                        "table": True
                    }
        self._datasets.append({
                    "type":"serviceDataset",
                    "dataSource":self._datasource,
                    "outFields":["*"],
                    "groupByFields":[],
                    "orderByFields":[],
                    "statisticDefinitions":self._data_statistic_defintion,
                    "querySpatialRelationship":"esriSpatialRelIntersects",
                    "returnGeometry":False,
                    "clientSideStatistics":False,
                    "name":"main"
                })
        if self._mindata.value_type == "fixedvalue":
            self._datasets.append({
                    "type":"staticDataset",
                    "data":self._mindata.min_value,
                    "name":"min"
                })

        elif self.mindata.value_type in ["statistic", "feature"]:
            if self.mindata.value_type == "statistic":
                self._mindata_statistic_defintion.append({
                            "onStatisticField":self._mindata.statistics_field,
                            "outStatisticFieldName":"value",
                            "statisticType":self._mindata.statistic
                        })
            self._datasets.append({
                    "type":"serviceDataset",
                    "dataSource":{
                        "type":"featureServiceDataSource",
                        "itemId":self.item.itemid,
                        "layerId":0,
                        "table":False
                    },
                    "outFields":["*"],
                    "groupByFields":[],
                    "orderByFields":[],
                    "statisticDefinitions":self._mindata_statistic_defintion,
                    "querySpatialRelationship":"esriSpatialRelIntersects",
                    "returnGeometry":False,
                    "clientSideStatistics":False,
                    "name":"min"
                })

        if self._maxdata.value_type == "fixedvalue":
            self._datasets.append({
                    "type":"staticDataset",
                    "data":self._maxdata.max_value,
                    "name":"max"
                })
        elif self.maxdata.value_type in ["statistic", "feature"]:
            if self.maxdata.value_type == "statistic":
                self._maxdata_statistic_defintion.append({
                            "onStatisticField":self._maxdata.statistics_field,
                            "outStatisticFieldName":"value",
                            "statisticType":self._maxdata.statistic
                        })
            self._datasets.append({
                    "type":"serviceDataset",
                    "dataSource":{
                        "type":"featureServiceDataSource",
                        "itemId":self.item.itemid,
                        "layerId":0,
                        "table":False
                    },
                    "outFields":["*"],
                    "groupByFields":[],
                    "orderByFields":[],
                    "statisticDefinitions":self._maxdata_statistic_defintion,
                    "querySpatialRelationship":"esriSpatialRelIntersects",
                    "returnGeometry":False,
                    "clientSideStatistics":False,
                    "name":"max"
                })
        json_data = {
            "type":"gaugeWidget",
            "style":self._gauge.style,
            "displayAsPercentage":False,
            "valueConversion":{
                "factor":self._data.factor,
                "offset":self._data.offset
            },
            "minimumConversion":{
                "factor":self._mindata.factor,
                "offset":self._mindata.offset
            },
            "maximumConversion":{
                "factor":self._maxdata.factor,
                "offset":self._maxdata.offset
            },
            "valueFormat":{
                "name":"value",
                "type":"decimal",
                "prefix":True,
                "pattern":"#,###.#"
            },
            "percentageFormat":{
                "name":"percentage",
                "type":"decimal",
                "prefix":False,
                "pattern":"#.#%"
            },
            "arrows":self._arrows,
            "axes":self._axes,
            "labels":[
                {
                    "id":"value",
                    "align":"center",
                    "color":"#ff0000",
                    "size":12,
                    "y":"40%"
                }
            ],
            "valueField":self._data.value_field,
            "noValueVerticalAlignment":self._novalue.alignment,
            "showCaptionWhenNoValue":self._novalue.show_title,
            "showDescriptionWhenNoValue":self._novalue.show_description,
            "valueType":self._data.value_type,
            "backgroundColor":self._background_color,
            "textColor":self._text_color,
            "datasets":self._datasets,
            "id":self._id,
            "name":self.name,
            "caption":self.title,
            "description":self.description,
            "showLastUpdate":self.show_last_update,
            "noDataText":self._nodata.text,
            "noDataVerticalAlignment":self._nodata.alignment,
            "showCaptionWhenNoData":self._nodata.show_title,
            "showDescriptionWhenNoData":self._nodata.show_description
        }            
        return json_data


class GaugeData(object):

    @classmethod
    def _create_data(cls, data_item=None, value_type="statistic"):
        data = GaugeData()

        data._item = data_item

        data._value_type = "statistic"
        data._min_value = 1
        data._max_value = 100
        data._statistic = "count"
        data._value_field = data.objectid_field
        data._statistics_field = data.objectid_field
        data._factor = 1
        data._offset = 0
        data._filters = []

        data.value_type = value_type

        return data

    @property
    def value_type(self):
        """
        :return: value type of data.
        """
        return self._value_type

    @value_type.setter
    def value_type(self, value):
        """
        Set value type of data.
        """
        if value in ["fixedvalue", "statistic", "feature"]:
            self._value_type = str(value)
        else:
            raise Exception("Select correct value type. Supported value types are 'fixedvalue', 'statistic', 'feature'")
    
    @property
    def min_value(self):
        """
        :return: min value of data.
        """
        return self._min_value

    @min_value.setter
    def min_value(self, value):
        """
        Set min value of data.
        """
        if self.value_type == "fixedvalue":
            self._min_value = value
        else:
            raise Exception("Can set minimum value for only 'fixedvalue' value type")

    @property
    def max_value(self):
        """
        :return: max value of data.
        """
        return self._max_value

    @max_value.setter
    def max_value(self, value):
        """
        Set max vaue of data.
        """
        if self.value_type == "fixedvalue":
            self._max_value = value
        else:
            raise Exception("Can set maximum value for only 'fixedvalue' value type")

    @property
    def statistic(self):
        """
        :return: statistic for data.
        """
        return self._statistic

    @statistic.setter
    def statistic(self, value):
        """
        Set statistic for data.
        """
        if self._value_type == "statistic":
            if value in ['count', 'avg', 'min', 'max', 'stddev', 'sum']:
                self._statistic = value
        else:
            raise Exception("Can set statistic only for reference type 'statistic'")

    @property
    def factor(self):
        """
        :return: value conversion factor.
        """
        return self._factor

    @factor.setter
    def factor(self, value):
        """
        Set value conversion factor.
        """
        self._factor = value

    @property
    def offset(self):
        """
        :return: value conversion offset.
        """
        return self._offset

    @offset.setter
    def offset(self, value):
        """
        Set value conversion offset.
        """
        self._offset = value
    
    @property
    def objectid_field(self):
        """
        :return: object ID field name.
        """
        return self._item.tables[0].properties["objectIdField"]
    
    @property
    def statistics_field(self):
        """
        :return: Statistics field name.
        """
        return self._statistics_field
        
    @property
    def value_field(self):
        """
        :return: value field for data.
        """
        return self._value_field

    @value_field.setter
    def value_field(self, value):
        """
        Set value field for data.
        """
        if self._value_type == "feature":
            f_type = self._field_type(value)
            if f_type in ["float64", "float32", "int64", "int32"]:
                self._value_field = value
            else:
                raise Exception("Please select a numeric field")
        else:
            raise Exception("Can add value field only for value type 'feature'")
    
    @property
    def filters(self):
        """
        :return: filters associated with widget
        """
        return self._filters
    
    def add_filter(self, field, join, condition, **kwargs):
        """
        Add filters associated with widget.
        """
        self._filter_field = field
        if join in ["AND", "OR"]:
            self._filter_join = join
        else:
            raise Exception("Please select from 'AND', 'OR'")
        if condition in ["between", "not between", "equal", "not equal", "greater than", "greater than or equal", "less than", "less than or equal", "is null", "is not null"]:
            self._filter_condition = condition
        else:
            raise Exception("Please select the right condition")

        if condition in ["between", "not between"]:
            self._val1 = kwargs.get('start')
            self._val2 = kwargs.get('end')
            self._filters.append({"filtertype":self._filter_join, "field":self._filter_field, "operator":self._filter_condition, "start":self._val1, "end":self._val2})
        else:
            raise Exception("Please provide 'start' and 'end' values as parameters")

        if condition in ["equal", "not equal", "greater than", "greater than or equal", "less than", "less than or equal"]:
            self._val = kwargs.get('value')
            self._filters.append({"filtertype":self._filter_join, "field":self._filter_field, "operator":self._filter_condition, "value":self._val})
        else:
            raise Exception("Please provide a 'value' parameter for comparison")
         
    def _field_type(self, field_name):
        f_type = self._item.tables[0].query().sdf[field_name].dtype
        return f_type
    

class GaugeProperties(object):

    @classmethod
    def _gauge_init(cls, style="progress", shape="horseshoe"):
        gauge = GaugeProperties()

        gauge._style = "progress"
        gauge._shape = "horseshoe"

        gauge.style = style
        gauge.shape = shape

        return gauge

    @property
    def style(self):
        """
        :return: gauge style.
        """
        return self._style

    @style.setter
    def style(self, value):
        """
        Set gauge style.
        """
        if value in ["progress", "meter"]:
            self._style = str(value)
        else:
            raise Exception("Please select gauge style from 'progress', 'meter'")

    @property
    def shape(self):
        """
        :return: gauge shape.
        """
        return self._shape

    @shape.setter
    def shape(self, value):
        """
        Set gauge shape.
        """
        if value in ["horseshoe", "halfdonut"]:
            self._shape = str(value)
        elif value == "circle":
            if self._style == "progress":
                self._shape = str(value)
            else:
                raise Exception("'circle' shape can only be selected for style 'progress'")
        else:
            raise Exception("Please select gauge shape from 'circle', 'horseshoe', 'halfdonut'")




import math
from itertools import compress
from .env import HAS_TENSORFLOW
if HAS_TENSORFLOW:
    import tensorflow as tf
    from tensorflow.keras.layers import Input, Conv2D, Dropout, ReLU, BatchNormalization, \
                                        UpSampling2D, Reshape, Layer, AveragePooling2D, \
                                        MaxPool2D, GlobalAveragePooling2D, GlobalMaxPool2D, \
                                        Concatenate, Flatten, Dense
    from tensorflow.keras import Model
    from .._utils.fastai_tf_fit import _tf_to_pytorch, _pytorch_to_tf_batch, _pytorch_to_tf
    from .common_tf import NormalizationLayerRGB

from .common import get_nbatches, image_batch_stretcher

try:
    import torch
    import fastai
    import numpy as np
    import matplotlib.pyplot as plt
    from fastai.vision import imagenet_stats
    HAS_FASTAI = True
except:
    HAS_FASTAI = False

## Common section starts

def IC_show_results(self, nrows=5, **kwargs):
    
    type_data_loader = kwargs.get('data_loader', 'validation') # options : traininig, validation, testing
    if type_data_loader == 'training':
        data_loader = self._data.train_dl
    elif type_data_loader == 'validation':
        data_loader = self._data.valid_dl
    elif type_data_loader == 'testing':
        data_loader = self._data.test_dl
    else:
        e = Exception(f'could not find {type_data_loader} in data. Please ensure that the data loader type is traininig, validation or testing ')
        raise(e)

    thresh = kwargs.get('thresh', 0.5)
    nodata = kwargs.get('nodata', 0)
    index = kwargs.get('start_index', 0)
    imsize = kwargs.get('imsize', 5)
    statistics_type = kwargs.get('statistics_type', 'dataset') # Accepted Values `dataset`, `DRA`
    stretch_type = kwargs.get('stretch_type', 'minmax') # Accepted Values `minmax`, `percentclip`

    # Get Batch
    n_items = min(nrows, len(data_loader.x))
    nbatches = math.ceil(n_items/self._data.batch_size)
    x_batch, y_batch = get_nbatches(data_loader, nbatches)
    x_batch = torch.cat(x_batch)
    y_batch = torch.cat(y_batch)

    # Get Predictions
    predictions_class_store = []
    predictions = []
    for i in range(0, x_batch.shape[0], self._data.batch_size):
        if self._backend == 'pytorch':
            batch_preds = self.learn.pred_batch(batch=(x_batch[i:i+self._data.batch_size], y_batch[i:i+self._data.batch_size]))
            # batch_preds: torch.tensor(B,C), where B is the batch size and C is the number of classes
            if self._data.dataset_type == "MultiLabeled_Tiles":
                predictions.append(batch_preds)
            else:
                confidences, class_idxs = torch.max(batch_preds, dim=1)
                predictions_class_store.extend(class_idxs)

        elif self._backend == 'tensorflow':
            class_idxs, _confidences = TF_IC_predict_batch(self, x_batch[i:i+self._data.batch_size])
            predictions_class_store.extend(class_idxs)
    
    # predictions will only hold values with Multilabel_Tiles
    # convert predictions from List[torch.tensor] to a torch.tensor
    if predictions: predictions = torch.cat(predictions)

    if self._is_multispectral:

        rgb_bands = kwargs.get('rgb_bands', self._data._symbology_rgb_bands)

        e = Exception('`rgb_bands` should be a valid band_order, list or tuple of length 3 or 1.')
        symbology_bands = []
        if not ( len(rgb_bands) == 3 or len(rgb_bands) == 1 ):
            raise(e)
        for b in rgb_bands:
            if type(b) == str:
                b_index = self._bands.index(b)
            elif type(b) == int:
                self._bands[b] # To check if the band index specified by the user really exists.
                b_index = b
            else:
                raise(e)
            b_index = self._data._extract_bands.index(b_index)
            symbology_bands.append(b_index)

        # Denormalize X
        x_batch = (self._data._scaled_std_values[self._data._extract_bands].view(1, -1, 1, 1).to(x_batch) * x_batch ) + self._data._scaled_mean_values[self._data._extract_bands].view(1, -1, 1, 1).to(x_batch)
        
        # Extract RGB Bands
        symbology_x_batch = x_batch[:, symbology_bands]
        if stretch_type is not None:
            symbology_x_batch = image_batch_stretcher(symbology_x_batch, stretch_type, statistics_type)

    else:
        # normalization stats
        norm_mean = torch.tensor(imagenet_stats[0]).to(x_batch).view(1, -1, 1, 1)
        norm_std = torch.tensor(imagenet_stats[1]).to(x_batch).view(1, -1, 1, 1)
        symbology_x_batch = (x_batch * norm_std) + norm_mean

    # Channel first to channel last for plotting
    symbology_x_batch = symbology_x_batch.permute(0, 2, 3, 1)

    # Clamp float values to range 0 - 1
    if symbology_x_batch.mean() < 1:
        symbology_x_batch = symbology_x_batch.clamp(0, 1)

    # Squeeze channels if single channel (1, 224, 224) -> (224, 224)
    if symbology_x_batch.shape[-1] == 1:
        symbology_x_batch = symbology_x_batch.squeeze()

    # Get color Array
    color_array = self._data._multispectral_color_array

    # Handle Sparse Data (not to be applied with Multilabel data)
    if self._data.dataset_type != "MultiLabeled_Tiles" and y_batch.ndim > 1:
        y_batch = y_batch.max(-1)[1]
    
    # Plotting Ground Truth and Prediction side by side
    ncols = 2
    title_font_size = 16
    _top = 1 - (math.sqrt(title_font_size)/math.sqrt(100*n_items*imsize))
    top = kwargs.get('top', _top)
    fig, axs = plt.subplots(nrows=n_items, ncols=ncols, figsize=(ncols*imsize, n_items*imsize))
    fig.suptitle('Ground truth/Predictions', fontsize=title_font_size, weight='bold')
    plt.subplots_adjust(top=top)
    idx=0
    for r in range(n_items):
        if n_items==1:
            ax_i = axs
        else:
            ax_i  = axs[r]

        # Get ground truth and prediction class names
        if self._data.dataset_type == "MultiLabeled_Tiles":
            one_hot_labels = y_batch[idx].tolist()
            gt_label = ";".join(compress(self._data.classes, one_hot_labels))
            one_hot_pred = (predictions[idx] >= thresh).tolist()
            prediction = ";".join(compress(self._data.classes, one_hot_pred))
        #For single label (Pytorch and TF)
        else:
            gt_label = self._data.classes[y_batch[idx].item()]
            prediction = self._data.classes[predictions_class_store[idx]]

        #Plot ground truth
        ax_ground_truth = ax_i[0]
        ax_ground_truth.axis('off')
        ax_ground_truth.imshow(symbology_x_batch[idx].cpu().numpy())
        ax_ground_truth.set_title(gt_label)
            
        # Plot Predictions
        ax_prediction  = ax_i[1]
        ax_prediction.axis('off')
        ax_prediction.imshow(symbology_x_batch[idx].cpu().numpy())
        ax_prediction.set_title(prediction)
        
        idx+=1
    return fig, axs

## Common section ends


## Tensorflow specific utils start ##

def TF_IC_predict_batch(self, imagetensor_batch):
    predictions = self.learn.model(_pytorch_to_tf_batch(imagetensor_batch)).detach().numpy()
    predictions_conf = (predictions.max(axis=-1)*100).tolist()
    predicted_classes = predictions.argmax(axis=-1).tolist()
    return predicted_classes, predictions_conf


def TF_IC_get_head_output(arcgis_model_fc):
    backbone_name = arcgis_model_fc._backbone_initalized.name
    if 'resnet' in backbone_name and any(x in backbone_name for x in ['50', '101', '151']):
        pyramid_scheme = [2048, 512]
    else:
        pyramid_scheme = [512]
    
    data_format = tf.keras.backend.image_data_format() #'channels_first' or 'channels_last'
    if data_format == 'channels_last':
        channel_axis = -1
    else:
        channel_axis = 1

    ## FC Head

    #Zero Block
    input_layer = arcgis_model_fc._backbone_initalized.output
    
    # First Block
    in_spatial_size = input_layer.shape[1]
    # head_block1_avgpool1 = AveragePooling2D(pool_size=(in_spatial_size, in_spatial_size), name="head_block1_globalavgpool1")(input_layer)
    # head_block1_avgpool1 = Flatten()(head_block1_avgpool1)
    head_block1_avgpool1 = GlobalAveragePooling2D(name="head_block1_globalavgpool1")(input_layer) 

    # Because tflite (tensorflow 2.0.0) does not support REDUCE_MAX operation MaxPool2D is used and the output is 
    # flatten that instead of directly using GlobalMaxPool2D
    #head_block1_maxpool1 = GlobalMaxPool2D(name="head_block1_globalmaxpool1")(input_layer)
    head_block1_maxpool1 = MaxPool2D(pool_size=(in_spatial_size, in_spatial_size), name="head_block1_globalmaxpool1")(input_layer)
    head_block1_maxpool1 = Flatten()(head_block1_maxpool1)

    head_block1_concat1 = Concatenate(axis=channel_axis, name="head_block1_concat1")([head_block1_avgpool1, head_block1_maxpool1])


    # Out layer for further use
    out_layer = head_block1_concat1

    if arcgis_model_fc._fpn:
        # Fpn Block
        for pyramid_size  in pyramid_scheme:
            fpn_dense = Dense(pyramid_size, activation='relu', name=f"head_fpnblock_dense_{pyramid_size}")(out_layer)
            fpn_bn = BatchNormalization(axis=channel_axis, name=f"head_fpnblock_bn_{pyramid_size}")(fpn_dense)
            out_layer = fpn_bn

    # Final Block
    head_final_drop1 = Dropout(0.3, name="head_final_drop1")(out_layer)
    head_final_logits_dense1 = Dense(arcgis_model_fc._data.c, name="head_final_logits_dense1")(head_final_drop1)
    
    return head_final_logits_dense1


def get_TFIC_post_processed_model(arcgis_model, input_normalization=True):
    model = arcgis_model.learn.model
    input_layer = model.input
    model_output = model.output

    if input_normalization:
        input_layer = Input(tuple(input_layer.shape[1:]))
        x = NormalizationLayerRGB()(input_layer)
        model_output = model(x)
    output_layer = tf.nn.softmax(model_output, axis=-1)
    new_model = Model(input_layer, output_layer)
    return new_model

## Tensorflow specific utils end ##



from ._codetemplate import super_resolution
import json
import traceback
from ._arcgis_model import _EmptyData
from .._data import _raise_fastai_import_error  
try:
    from ._arcgis_model import ArcGISModel, _resnet_family
    from ._superres_utils import FeatureLoss, gram_matrix, compute_metrics, get_resize, create_loss
    from fastai.vision.learner import unet_learner
    from fastai.vision import nn, ImageImageList, get_transforms, imagenet_stats, NormType, open_image
    from fastai.callbacks import LossMetrics
    from fastai.utils.mem import Path
    from .._utils.common import _get_emd_path

    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_FASTAI = False




class SuperResolution(ArcGISModel):

    """
    Creates a model object which increases the resolution and improves the quality of images.
    Based on Fast.ai MOOC Lesson 7.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            `prepare_data` function.
    ---------------------   -------------------------------------------
    backbone                Optional function. Backbone CNN model to be used for
                            creating the base of the `UnetClassifier`, which
                            is `resnet34` by default.
                            Compatible backbones: 'resnet18', 'resnet34', 'resnet50', 'resnet101', 'resnet152'
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    =====================   ===========================================
                                             
    :returns: `SuperResolution` Object
    """
    def __init__(self, data, backbone=None, pretrained_path=None, *args, **kwargs):
        super().__init__(data, backbone, **kwargs)
        self._check_dataset_support(data)
        feat_loss = create_loss(self._device.type)
        data.c = 3
        self.learn = unet_learner(data, arch=self._backbone, wd=1e-3, loss_func=feat_loss, callback_fns=LossMetrics, blur=True, norm_type=NormType.Weight)
        self.learn.model = self.learn.model.to(self._device)
        if pretrained_path is not None:
            self.load(pretrained_path)
        
        self._code = super_resolution

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<%s>' % (type(self).__name__)

    @property
    def supported_backbones(self):
        """
        Supported torchvision backbones for this model.
        """        
        return SuperResolution._supported_backbones()

    @staticmethod
    def _supported_backbones():
        return [*_resnet_family]

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a SuperResolution object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_data` function or None for
                                inferencing.
        =====================   ===========================================
        
        :returns: `SuperResolution` Object
        """
        return cls.from_emd(data, emd_path)

    @classmethod
    def from_emd(cls, data, emd_path):
        """
        Creates a SuperResolution object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_data` function or None for
                                inferencing.
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Esri Model Definition
                                file.
        =====================   ===========================================
        
        :returns: `SuperResolution` Object
        """

        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)
            
        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd['ModelFile'])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        model_params = emd['ModelParameters']
        downsample_factor = emd.get('downsample_factor')
        resize_to = emd.get('resize_to')
        chip_size = emd['ImageHeight']
        feat_loss = create_loss()
        if data is None:
            data = (ImageImageList.from_folder(emd_path.parent.parent).split_none().label_from_func(lambda x: x).transform(get_transforms(do_flip=False),size=(chip_size, chip_size), tfm_y=True).databunch(bs=2, no_check=True).normalize(imagenet_stats, do_y=True))
            data._is_empty = True
            data.emd_path = emd_path
            data.downsample_factor = downsample_factor
            data.emd = emd
        data.resize_to = resize_to
        
        return cls(data, **model_params, pretrained_path=str(model_file))
    
    @property
    def _model_metrics(self):
        psnr_ssim = self.compute_metrics(show_progress=True)
        return {'psnr_metric': '{0:1.4e}'.format(psnr_ssim[0]),
                'ssim_metric': '{0:1.4e}'.format(psnr_ssim[1])}

    def _get_emd_params(self, save_inference_file):
        _emd_template = {}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        _emd_template["ModelConfiguration"] = "_superres"
        _emd_template["InferenceFunction"] = "ArcGISSuperResolution.py"
        _emd_template["ModelType"] = "SuperResolution"
        _emd_template["downsample_factor"] = self._data.downsample_factor
        return _emd_template

    def compute_metrics(self, accuracy=True, show_progress=True):
        """
        Computes Peak Signal-to-Noise Ratio (PSNR) and 
        Structural Similarity Index Measure (SSIM) on validation set.

        """
        psnr, ssim = compute_metrics(self, self._data.valid_dl, show_progress)
        return psnr, ssim

    
    def show_results(self, rows=5):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        =====================   ===========================================

        """
        if rows > len(self._data.valid_ds):
            rows = len(self._data.valid_ds)
        
        self._check_requisites()
        self.learn.show_results(rows=rows)

    def predict(self, img_path, width=None, height=None):
        """
        Predicts and display the image.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        img_path                Required path of an image.
        ---------------------   -------------------------------------------
        width                   Optional int. Width of the predicted 
                                output image.
        ---------------------   -------------------------------------------
        height                  Optional int. Height of the predicted
                                output image.
        =====================   ===========================================

        """
        img_path = Path(img_path)
        img = open_image(img_path)        
        temp_databunch = self.learn.data
        if width is not None or height is not None:
            if width is None:
                width = height
            elif height is None:
                height = width
        elif width is None and height is None:
            _,width,height = img.shape
        
        y_new, z_new = width, height
        pred_databunch = (ImageImageList.from_folder(img_path.parent).split_none()\
        .label_from_func(lambda x: x)\
        .transform(get_transforms(do_flip=False), size=(height,width), tfm_y=True)\
        .databunch(bs=2, no_check=True).normalize(imagenet_stats, do_y=True))
            
        self.learn.data = pred_databunch
        
        pred_img = self.learn.predict(img)[0]
        self.learn.data = temp_databunch
        return pred_img

    @property
    def  supported_datasets(self):
        """ Supported dataset types for this model. """
        return SuperResolution._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ['Export_Tiles', 'superres'] 






import traceback
from .._data import _raise_fastai_import_error
HAS_TRANSFORMER = True

try:
    import torch
    from transformers import pipeline, logging
    from .._utils.common import _get_device_id
    from fastprogress.fastprogress import progress_bar
    from transformers import AutoTokenizer, AutoModelForSeq2SeqLM
except Exception as e:
    transformer_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_TRANSFORMER = False


class TextTranslator:
    """
    Creates a `TextTranslator` Object.
    Based on the Hugging Face transformers library
    To learn more about the available models for translation task,
    kindly visit:- https://huggingface.co/models?filter=translation
    
    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    source_language         Optional string. Specify the language of the
                            text you would like to get the translation of.
                            Default value is 'es' (Spanish)
    ---------------------   -------------------------------------------
    target_language         Optional string. The language into which one
                            wishes to translate the input text.
                            Default value is 'en' (English)
    =====================   ===========================================

    :returns: `TextTranslator` Object
    """

    #: supported transformer backbones
    supported_backbones = ["MarianMT"]

    def __init__(self, source_language="es", target_language="en"):
        if not HAS_TRANSFORMER:
            _raise_fastai_import_error(import_exception=transformer_exception)

        logger = logging.get_logger()
        logger.setLevel(logging.ERROR)
        self._source_lang = source_language
        self._target_lang = target_language

        self._device_id = _get_device_id()
        self._device = torch.device("cpu" if self._device_id < 0 else "cuda:{}".format(self._device_id))
        self._task = f"translation_{self._source_lang}_to_{self._target_lang}"
        self._tokenizer = AutoTokenizer.from_pretrained(f"Helsinki-NLP/opus-mt-{source_language}-{target_language}")
        self.model = AutoModelForSeq2SeqLM.from_pretrained(f"Helsinki-NLP/opus-mt-{source_language}-{target_language}")
        self.model.to(self._device)

    def translate(self, text_or_list, **kwargs):
        """
        Translate the given text or list of text into the target language

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        text_or_list            Required string or list. A text/passage
                                or a list of texts/passages to translate.
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        min_length              Optional integer. The minimum length of the
                                sequence to be generated.
                                Default value is set to to `min_length` parameter
                                of the model config.
        ---------------------   -------------------------------------------
        max_length              Optional integer. The maximum length of the
                                sequence to be generated.
                                Default value is set to to `max_length` parameter
                                of the model config.
        ---------------------   -------------------------------------------
        num_return_sequences    Optional integer. The number of independently
                                computed returned sequences for each element
                                in the batch.
                                Default value is set to 1.
        ---------------------   -------------------------------------------
        num_beams               Optional integer. Number of beams for beam
                                search. 1 means no beam search.
                                Default value is set to 1.
        ---------------------   -------------------------------------------
        length_penalty          Optional float. Exponential penalty to the
                                length. 1.0 means no penalty. Set to values < 1.0
                                in order to encourage the model to generate
                                shorter sequences, to a value > 1.0 in order to
                                encourage the model to produce longer sequences.
                                Default value is set to 1.0.
        ---------------------   -------------------------------------------
        early_stopping          Optional bool. Whether to stop the beam search
                                when at least ``num_beams`` sentences are
                                finished per batch or not.
                                Default value is set to False.
        =====================   ===========================================

        :returns: a list or a list of list containing the translation of the input prompt(s) / sentence(s) to the target language
        """
        results = []
        num_return_sequences = kwargs.get("num_return_sequences", 1)
        if not isinstance(text_or_list, (list, tuple)): text_or_list = [text_or_list]
        for i in progress_bar(range(len(text_or_list))):
            inputs = self._tokenizer.encode(f"{text_or_list[i]} {self._tokenizer.eos_token}", return_tensors="pt").\
                to(self._device)
            outputs = self.model.generate(inputs, **kwargs)
            result = self._tokenizer.batch_decode(outputs, skip_special_tokens=True)
            result = [{"translated_text": x} for x in result]
            if num_return_sequences == 1: result = result[0]
            results.append(result)
        return results



from arcgis.gis import GIS
from arcgis.env import active_gis
from arcgis._impl.common._mixins import PropertyMap
from arcgis.features._version import Version

###########################################################################
class ValidationManager(object):
    """
    The Validation Server is responsible for exposing the management 
    capabilities necessary to support evaluation of geodatabase rules.
    """
    _con = None
    _gis = None
    _url = None
    _version = None
    _properties = None
    
    #----------------------------------------------------------------------
    def __init__(self, url, version=None, gis=None):
        """initializer"""
        self._url = url
        if gis is None:
            self._gis = active_gis
        else:
            self._gis = gis
        assert isinstance(self._gis, GIS)
        self._con = gis._con
        self._version = version
        assert isinstance(self._version, Version)
    #----------------------------------------------------------------------
    def _init(self):
        """loads the properties"""
        try:
            self._properties = PropertyMap(self._con.get(self._url, {'f' : 'json'}))
        except:
            self._properties = PropertyMap(self._con.post(self._url, {'f' : 'json'}))
    #----------------------------------------------------------------------
    def __str__(self):
        url = self._url
        return f"<ValidationManager @ {url}>"
    #----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()    
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """service properties"""
        if self._properties is None:
            self._init()
        return self._properties
    #----------------------------------------------------------------------
    def update_error(self, 
                     error_features, 
                     version=None, 
                     return_edits=None,
                     **kwargs):
        """
        Updates errors on the validation tables.  
        
        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        error_features      Required List.  The error features to be updated. 
                            
                            **Syntax**
                            
                            ```
                            [
                                {
                                  "errorType" : "object" | "point" | "line" |
                                               "polygon",
                                  "features" : [
                                    {
                                      "globalId" : <guid>,
                                      "fields" : {
                                        "name1" : <value1>,
                                        "name2" : <value2> 
                                      }
                                    }
                                  ]
                                }  
                            ]
                            ```
        ---------------     --------------------------------------------------------------------
        return_edits        Optional Boolean. `return_edits` returns features edited due to 
                            errors update. Results returned are organized in a layer by layer 
                            fashion. If it is set to `True`, each layer may have edited features 
                            returned in an editedFeatures object.

                            The editedFeatures object returns full features including the original 
                            features prior to delete, the original and current features for 
                            updates and the current rows for inserts which may contain implicit 
                            changes (e.g. as a result of a calculation rule ).

                            The response includes no editedFeatures and `exceededTransferLimit=True`
                            if the count of edited features to return is more than the maxRecordCount. 
                            If clients are using this parameter to maintain a cache, they should 
                            invalidate the cache when `exceededTransferLimit = True` is returned.
                            If the server encounters an error when generating the list of edits is 
                            the response, `exceededTransferLimit = True` is also returned.

                            Edited features are returned in the spatial reference of the feature 
                            service as defined by the services spatialReference object or by the
                            spatialReference of the layers extent object.

                            The default for this parameter is False.
        ===============     ====================================================================
        
        :returns: dictionary
        
        """
        url = self._url + "/updateErrors"
        
        version = self._version
        version_name = version.properties.versionName
        session_id = version._guid
        params = {
            'f' : 'json',
            'gdbVersion' : version_name,
            'sessionId' : session_id,
            'errorFeatures' : error_features
        }
        if len(kwargs) > 0:
            params.update(kwargs)
        if not return_edits is None:
            params['returnEdits'] = return_edits
        res = self._con.post(url, params)
        return res
    #----------------------------------------------------------------------
    def evaluate(self, 
                 evaluation,
                 area=None, 
                 changes_in_version=False,
                 selection=None,
                 return_edits=False):
        """
        Runs the topology rules and returns new errors if they exist.
        
        ====================     ====================================================================
        **Argument**             **Description**
        --------------------     --------------------------------------------------------------------
        evaluation               Required List of Strings.  An array of evaluation types.
                                 Allowed Rule Types: `validation`, `calculation` and/or `topology`
        --------------------     --------------------------------------------------------------------
        area                     Optional Envelope/Dict. Extent area to evaluate. 
        --------------------     --------------------------------------------------------------------
        changes_in_version       Optional Boolean. representing whether to perform the evaluation on 
                                 the features that have changed in the version (default is false). 
                                 Does not apply to the DEFAULT version.
                                 
                                 When set to true, the evaluationDate property for the version is 
                                 updated. This is listed as a property for a version and can be 
                                 accessed using the version resource and the version infos operation.
        --------------------     --------------------------------------------------------------------
        selection                Optional List.  A set of features to evaluate.  This is an array of 
                                 layers and the global IDs or Object IDs of the features to examine.
                                 
                                 If the `evaluation_type` is **topology** this parameter is ignored.
                                 
                                 **Syntax**
                                 
                                 ```
                                 [ 
                                    {
                                      "id" : <layerId1>,
                                      "globalIds" : [ <globalId> ],
                                      "objectIds" : [ <objectId> ]
                                    },
                                    {
                                      "id" : <layerId2>,
                                      "globalIds" : [ <globalId> ].
                                      "objectIds" : [ <objectId> ]
                                    }
                                 ]
                                 ```
        
        --------------------     --------------------------------------------------------------------
        return_edits             Optional Boolean. returns features edited due to feature evaluation. 
                                 Results returned are organized in a layer by layer fashion. If 
                                 `return_edits` is set to true, each layer may have edited features 
                                 returned.
                            
                                 The default for this parameter is false. Always set to true when 
                                 evaluating `topology` for a parcel fabric.
                                 
        ====================     ====================================================================
        
        :returns: dictionary
        
        """
        url = self._url + "/evaluate"
        eval_lu = {
            'validation' : "validationRules",
            "validationrules" : "validationRules",
            "calculation" : "calculationRules",
            "calculationrules" : "calculationRules",
            "topology" : "topologyRules",
            "topologyrules" : "topologyRules"
                           
        }
        params = {
            'f' : 'json',
            'sessionId' : self._version._guid,
            'gdbVersion' : self._version.properties.versionName,
            'changesInVersion' : changes_in_version,
            'returnEdits' : return_edits,
            'async' : False
        }
        if not area is None:
            params['evaluationArea'] = area
        if not selection is None:
            params['selection'] = selection
            
        if isinstance(evaluation, str) and \
           evaluation.lower() in eval_lu:
            params['evaluationType'] = [eval_lu[evaluation.lower()]]
        elif isinstance(evaluation, str) and \
             not evaluation.lower() in eval_lu:
            raise ValueError("Invalid evalution type")
        elif isinstance(evaluation, (list, tuple)):
            params['evaluationType'] = [eval_lu[e.lower()] for e in evaluation]
        else:
            raise ValueError("Invalid evalution type")
        
        return self._con.post(url, params)
            
            


from ._codetemplate import image_translation_prf
import json
import traceback
from .._data import _raise_fastai_import_error
from ._arcgis_model import ArcGISModel
import logging
logger = logging.getLogger()
try:
    from ._cyclegan_utils import CycleGanLoss, CycleGANTrainer, optim, compute_fid_metric
    from ._cyclegan_utils import  CycleGAN as CycleGAN_model
    from .._utils.cyclegan import ImageTuple, ImageTupleList, ImageTupleListMS
    from .._utils.common import get_multispectral_data_params_from_emd, _get_emd_path, ArcGISMSImage
    from torchvision import transforms
    from pathlib import Path
    from fastai.vision import DatasetType, Learner, partial, open_image, Image
    import torch

    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_FASTAI = False

class CycleGAN(ArcGISModel):

    """
    Creates a model object which generates images of type A from type B or type B from type A.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            `prepare_data` function.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    ---------------------   -------------------------------------------
    gen_blocks              Optional integer. Number of ResNet blocks to use 
                            in generator.
    ---------------------   -------------------------------------------
    lsgan                   Optional boolean. If True, it will use Mean Squared Error
                            else it will use Binary Cross Entropy.
    =====================   ===========================================
                                             
    :returns: `CycleGAN` Object
    """
    def __init__(self, data, pretrained_path=None, gen_blocks=9, lsgan=True, *args, **kwargs):
        super().__init__(data)
        cycle_gan = CycleGAN_model(self._data.n_channel,self._data.n_channel, gen_blocks=gen_blocks, lsgan=lsgan)
        self.learn = Learner(data, cycle_gan, loss_func=CycleGanLoss(cycle_gan), opt_func=partial(optim.Adam, betas=(0.5,0.99)),callback_fns=[CycleGANTrainer])
        self.learn.model = self.learn.model.to(self._device)
        self._slice_lr = False
        if pretrained_path is not None:
            self.load(pretrained_path)
        self._code = image_translation_prf
        def __str__(self):
            return self.__repr__()
        def __repr__(self):
            return '<%s>' % (type(self).__name__)

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a CycleGAN object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_data` function or None for
                                inferencing.
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        =====================   ===========================================
        
        :returns: `CycleGAN` Object
        """

        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)
            
        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd['ModelFile'])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        model_params = emd['ModelParameters']
        resize_to = emd.get('resize_to')
        chip_size = emd['ImageHeight']
        if data is None:
            if emd.get('IsMultispectral', False):
                data = ImageTupleListMS.from_folders(emd_path.parent, emd_path.parent, emd_path.parent, batch_stats_a=None, batch_stats_b=None).split_none().label_empty().databunch(bs=2, no_check=True)
                data = get_multispectral_data_params_from_emd(data, emd)
                data._is_multispectral = emd.get('IsMultispectral', False)
                normalization_stats_b = dict(emd.get("NormalizationStats_b"))
                for _stat in normalization_stats_b:
                    if normalization_stats_b[_stat] is not None:
                        normalization_stats_b[_stat] = torch.tensor(normalization_stats_b[_stat])
                    setattr(data, ('_'+_stat), normalization_stats_b[_stat])

            else:
                data = ImageTupleList.from_folders(emd_path.parent, emd_path.parent, emd_path.parent)\
                    .split_none()\
                    .label_empty()\
                    .transform(size=(resize_to, resize_to))\
                    .databunch(bs=2, no_check=True)

            data.n_channel = emd['n_channel']
            data._is_empty = True
            data.emd_path = emd_path
            data.emd = emd

        data.resize_to = resize_to
        
        return cls(data, **model_params, pretrained_path=str(model_file))
        
    @property
    def _model_metrics(self):
        fid_a, fid_b = self.compute_metrics()
        return {'FID_A': f'{fid_a}', 'FID_B': f'{fid_b}'}

    def _get_emd_params(self, save_inference_file):
        _emd_template = {}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        _emd_template["ModelConfiguration"] = "_cyclegan"
        _emd_template["InferenceFunction"] = "ArcGISImageTranslation.py"
        _emd_template["ModelType"] = "CycleGAN"
        _emd_template["n_channel"] = self._data.n_channel
        _emd_template["SupportsVariableTileSize"] = True
        if self._data._is_multispectral:
            _emd_template["NormalizationStats_b"] = {
                    "band_min_values": self._data._band_min_values_b,
                    "band_max_values": self._data._band_max_values_b,
                    "band_mean_values": self._data._band_mean_values_b,
                    "band_std_values": self._data._band_std_values_b,
                    "scaled_min_values": self._data._scaled_min_values_b,
                    "scaled_max_values": self._data._scaled_max_values_b,
                    "scaled_mean_values": self._data._scaled_mean_values_b,
                    "scaled_std_values": self._data._scaled_std_values_b
        }
            for _stat in _emd_template["NormalizationStats_b"]:
                    if _emd_template["NormalizationStats_b"][_stat] is not None:
                        _emd_template["NormalizationStats_b"][_stat] = _emd_template["NormalizationStats_b"][_stat].tolist()
        return _emd_template

    def show_results(self, rows=5):
        """
        Displays the results of a trained model on a part of the validation set.

        """
        self.learn.model.arcgis_results = True
        self.learn.show_results()
        self.learn.model.arcgis_results = False

    def predict(self, img_path, convert_to):
        """
        Predicts and display the image.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        img_path                Required path of an image.
        ---------------------   -------------------------------------------
        convert_to              'A' if we want to generate image of type 'A' 
                                from type 'B' or 'B' if we want to generate 
                                image of type 'B' from type 'A' where A and
                                B are the domain specifications that were 
                                used while training.
        =====================   ===========================================

        """
        import numpy as np
        self.learn.model.arcgis_results = True
        img_path = Path(img_path)
        n_band = self._data.n_channel
        if self._data._is_multispectral:
            raw_img = ArcGISMSImage.open(img_path)
            if n_band > raw_img.shape[0]:
                cont = []
                last_tile = np.expand_dims(raw_img.data[raw_img.shape[0]-1,:,:], 0)
                res = abs(n_band - raw_img.shape[0])
                for i in range(res):
                    raw_img = Image(torch.tensor(np.concatenate((raw_img.data, last_tile), axis=0)))
        else:
            raw_img = open_image(img_path)
        raw_img_tuple = ImageTuple(raw_img, raw_img)
        pred_tuple = self.learn.predict(raw_img_tuple)
        if convert_to == 'A' or convert_to == 'a':
            pred_img = pred_tuple[1][0]/2+0.5
        elif convert_to == 'B' or convert_to == 'b':
            pred_img = pred_tuple[1][1]/2+0.5
        
        pred_img = transforms.ToPILImage()(pred_img).convert("RGB")
        self.learn.model.arcgis_results = False
        return pred_img

    def compute_metrics(self):
        """
        Computes Frechet Inception Distance (FID) on validation set.
        """
        if self._data._is_multispectral:
            logger.error("FID metric not supported for multispectral imagery type")
            return(None, None)
        else:
            fid_a, fid_b = compute_fid_metric(self, self._data)
            return fid_a, fid_b

        



from fastai.vision import ImageSegment, Image
from fastai.vision.image import open_image, show_image, pil2tensor
from fastai.vision.data import SegmentationProcessor, ImageList
from fastai.layers import CrossEntropyFlat
from fastai.basic_train import LearnerCallback
from fastai.core import is_listy
from .._utils.common import ArcGISMSImage, get_top_padding, kwarg_fill_none, \
    find_data_loader, get_nbatches, dynamic_range_adjustment, image_tensor_checks_plotting, \
    get_symbology_bands, predict_batch, denorm_x, get_nbatches, GDAL_INSTALL_MESSAGE, image_batch_stretcher
from .._utils.env import HAS_GDAL
from .._utils.pixel_classification import analyze_pred_pixel_classification
import torch
import warnings
import PIL
import numpy as np
import os
import math

def _class_array_to_rbg(ca : 'classified_array', cm : 'color_mapping', nodata=0):
    im = np.expand_dims(ca, axis=2).repeat(3, axis=2)
    white_mask = im[im == nodata]
    for x in (np.unique(im)):
        if not x == nodata:
            for i in range(3):
                im[:,:,i][im[:,:,i] == x] = cm[x][i]
    im[white_mask] = 255
    return im

#def _show_batch_unet_multispectral(self, nrows=3, ncols=3, n_items=None, index=0, rgb_bands=None, nodata=0, alpha=0.7, imsize=5): # Proposed Parameters
def _show_batch_unet_multispectral(self, rows=3, alpha=0.7, **kwargs): # parameters adjusted in kwargs
    import matplotlib.pyplot as plt
    from .._data import _tensor_scaler

    nrows = rows
    ncols = 3
    if kwargs.get('ncols', None) is not None:
        ncols = kwargs.get('ncols')

    n_items = None
    if kwargs.get('n_items', None) is not None:
        n_items = kwargs.get('n_items')

    type_data_loader = kwargs.get('data_loader', 'training') # options : traininig, validation, testing
    if type_data_loader == 'training':
        data_loader = self.train_dl
    elif type_data_loader == 'validation':
        data_loader = self.valid_dl
    elif type_data_loader == 'testing':
        data_loader = self.test_dl
    else:
        e = Exception(f'could not find {type_data_loader} in data. Please ensure that the data loader type is traininig, validation or testing ')
        raise(e)

    rgb_bands = self._symbology_rgb_bands
    if kwargs.get('rgb_bands', None) is not None:
        rgb_bands = kwargs.get('rgb_bands')

    nodata = 0
    if kwargs.get('nodata', None) is not None:
        nodata = kwargs.get('nodata')

    index = 0
    if kwargs.get('index', None) is not None:
        index = kwargs.get('index')

    imsize = 5
    if kwargs.get('imsize', None) is not None:
        imsize = kwargs.get('imsize')

    statistics_type = kwargs.get('statistics_type', 'dataset') # Accepted Values `dataset`, `DRA`
    stretch_type = kwargs.get('stretch_type', 'minmax') # Accepted Values `minmax`, `percentclip`

    e = Exception('`rgb_bands` should be a valid band_order, list or tuple of length 3 or 1.')
    symbology_bands = []
    if not ( len(rgb_bands) == 3 or len(rgb_bands) == 1 ):
        raise(e)
    for b in rgb_bands:
        if type(b) == str:
            b_index = self._bands.index(b)
        elif type(b) == int:
            self._bands[b] # To check if the band index specified by the user really exists.
            b_index = b
        else:
            raise(e)
        b_index = self._extract_bands.index(b_index)
        symbology_bands.append(b_index)

    # Get Batch
    if n_items is None:
        n_items = nrows*ncols
    else:
        nrows = math.ceil(n_items/ncols)
    n_items = min(n_items, len(self.x))

    x_batch, y_batch = get_nbatches(data_loader, n_items)
    x_batch = torch.cat(x_batch)
    y_batch = torch.cat(y_batch)
    # Denormalize X
    x_batch = (self._scaled_std_values[self._extract_bands].view(1, -1, 1, 1).to(x_batch) * x_batch ) + self._scaled_mean_values[self._extract_bands].view(1, -1, 1, 1).to(x_batch)


    # Extract RGB Bands
    symbology_x_batch = x_batch[:, symbology_bands]
    if stretch_type is not None:
        symbology_x_batch = image_batch_stretcher(symbology_x_batch, stretch_type, statistics_type)

    # Channel first to channel last and clamp float values to range 0 - 1 for plotting
    symbology_x_batch = symbology_x_batch.permute(0, 2, 3, 1)
    # Clamp float values to range 0 - 1
    if symbology_x_batch.mean() < 1:
        symbology_x_batch = symbology_x_batch.clamp(0, 1)

    # Squeeze channels if single channel (1, 224, 224) -> (224, 224)
    if symbology_x_batch.shape[-1] == 1:
        symbology_x_batch = symbology_x_batch.squeeze()

    # Get color Array
    color_array = self._multispectral_color_array
    color_array[1:, 3] = alpha

    # Size for plotting
    fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=(ncols*imsize, nrows*imsize))
    idx = 0
    for r in range(nrows):
        for c in range(ncols):
            if nrows == 1 and ncols == 1:
                axi = axs
            else:
                axi = axs[r][c]
            if idx < symbology_x_batch.shape[0]:
                axi.imshow(symbology_x_batch[idx])
                y_rgb = color_array[y_batch[idx][0]]#.cpu().numpy()
                #y_rgb = _class_array_to_rbg(y_batch[idx][0], self._multispectral_color_mapping, nodata)
                axi.imshow(y_rgb, alpha=alpha)
                axi.axis('off')
            else:
                axi.axis('off')
            idx+=1
    #

class ArcGISImageSegment(Image):
    "Support applying transforms to segmentation masks data in `px`."
    def __init__(self, x, color_mapping=None):
        super(ArcGISImageSegment, self).__init__(x)
        self.color_mapping = color_mapping

    def lighting(self, func, *args, **kwargs):
        return self

    def refresh(self):
        self.sample_kwargs['mode'] = 'nearest'
        return super().refresh()

    @property
    def data(self):
        "Return this image pixels as a `LongTensor`."
        return self.px.long()

    def show(self, ax=None, figsize:tuple=(3,3), title=None, hide_axis:bool=True,
        cmap=None, alpha:float=0.5, **kwargs):
        "Show the `ImageSegment` on `ax`."
        if is_no_color(self.color_mapping):
            ## This condition will not be true.
            ax = show_image(self, ax=ax, hide_axis=hide_axis, cmap="tab20", figsize=figsize,
                        interpolation='nearest', alpha=alpha, vmin=0, **kwargs)
        else:
            color_mapping = torch.tensor(list(self.color_mapping.values()))
            color_mapping = torch.cat((color_mapping.float()/255, torch.tensor([float(alpha)] * len(color_mapping)).view(-1, 1)), dim=1)
            color_mapping = torch.cat((torch.tensor([0., 0., 0., 0.]).view(1, -1), color_mapping), dim=0)
            try:
                color_im = color_mapping[self.data[0]].permute(2, 0, 1)
            except IndexError as e:

                if HAS_GDAL:
                    message = f"Encountered invalid values in training label values, please check your training data."
                else:
                    message = f"Encountered invalid values while reading training labels. Please install gdal for better support.\n\n" + GDAL_INSTALL_MESSAGE

                raise Exception(f"{e} \n\n{message}")

            ax = show_image(color_im, ax=ax, hide_axis=hide_axis, cmap=cmap, figsize=figsize,
                            interpolation='nearest', alpha=alpha, vmin=0, **kwargs)
        if title: ax.set_title(title)

class ArcGISMultispectralImageSegment():
    def __init__(self, tensor):
        self.data = tensor
        self.size = tensor.shape
        self.shape = tensor.shape

def is_no_color(color_mapping):
    if isinstance(color_mapping, dict):
        color_mapping = list(color_mapping.values())
    return (np.array(color_mapping) == [-1., -1., -1.]).any()

def is_contiguous(class_values):
    flag = True
    for i in range(len(class_values) - 1):
        if class_values[i] + 1 != class_values[i+1]:
            flag = False
    return flag

def map_to_contiguous(tensor, mapping):
    modified_tensor = torch.zeros_like(tensor)
    for i, value in enumerate(mapping):
        modified_tensor[tensor == value] = i
    return modified_tensor

class ArcGISSegmentationLabelList(ImageList):
    "`ItemList` for segmentation masks."
    _processor = SegmentationProcessor
    def __init__(self, items, classes=None, class_mapping=None, color_mapping=None, **kwargs):
        super().__init__(items, **kwargs)
        self.class_mapping = class_mapping
        self.color_mapping = color_mapping
        self.copy_new.append('classes')
        self.classes, self.loss_func = classes, CrossEntropyFlat(axis=1)
        self.is_contiguous = is_contiguous(sorted([0] + list(self.class_mapping.keys())))
        if not self.is_contiguous:
            self.pixel_mapping = [0] + list(self.class_mapping.keys())

    def _open_rgb(self, fn):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning) # EXIF warning from TiffPlugin
            x = PIL.Image.open(fn)
            if x.palette is not None:
                x = x.convert('P')
            else:
                x = x.convert('L')
            x = pil2tensor(x, np.float32)

        if not self.is_contiguous:
            x = map_to_contiguous(x, self.pixel_mapping)

        return ArcGISImageSegment(x, color_mapping=self.color_mapping)

    def analyze_pred(self, pred, thresh=0.5, ignore_mapped_class=[], model=None, thinning=None):

        if getattr(model, "_is_model_extension", False):

            if thinning is None:
                pred = model.model_conf.post_process(pred, thresh)
            else:
                pred = model.model_conf.post_process(pred, thresh, thinning)
            return pred

        if is_listy(pred):
            pred = pred[0]

        if ignore_mapped_class == []:
            return pred.argmax(dim=0)[None]
        else:
            for k in ignore_mapped_class:
                pred[k] = -1
            return pred.argmax(dim=0)[None]

    def reconstruct(self, t):
        return ArcGISImageSegment(t, color_mapping=self.color_mapping)

    def open(self, fn):
        x = ArcGISMSImage.open(fn).data
        if not self.is_contiguous:
            x = map_to_contiguous(x, self.pixel_mapping)
        return ArcGISImageSegment(x, color_mapping=self.color_mapping)

class ArcGISSegmentationItemList(ImageList):
    "`ItemList` suitable for segmentation tasks."
    _label_cls, _square_show_res = ArcGISSegmentationLabelList, False
    _div = None
    _imagery_type = None
    def open(self, fn):
        return ArcGISMSImage.open(fn, div=self._div, imagery_type=self._imagery_type)

class ArcGISSegmentationMSLabelList(ArcGISSegmentationLabelList):
    def open(self, fn):
        from osgeo import gdal
        path = str(os.path.abspath(fn))
        x = gdal.Open(path).ReadAsArray()
        x = torch.tensor(x.astype(np.float32))[None]
        if not self.is_contiguous:
            x = map_to_contiguous(x, self.pixel_mapping)
        return ArcGISImageSegment(x, color_mapping=self.color_mapping)

class LabelCallback(LearnerCallback):
    def __init__(self, learn):
        super().__init__(learn)
        self.label_mapping = {value:(idx+1) for idx, value in enumerate(learn.data.class_mapping.keys())}

    def on_batch_begin(self, last_input, last_target, **kwargs):
        """
        This callback is not used anymore.
        Not using this callback has increased the unet model training speed.
        """
        # modified_target = torch.zeros_like(last_target)
        # for label, idx in self.label_mapping.items():
        #     modified_target[last_target==label] = idx
        return {'last_input':last_input, 'last_target':last_target}


#def show_results_multispectral(self, nrows=3, index=0, type_ds='valid', rgb_bands=None, nodata=0, alpha=0.7, imsize=5, top=0.97): # Proposed Parameters
def show_results_multispectral(self, nrows=5, alpha=0.7, **kwargs): # parameters adjusted in kwargs
    import matplotlib.pyplot as plt

    # Get Number of items
    ncols = 2
    return_fig = kwargs.get('return_fig', False)
    type_data_loader = kwarg_fill_none(kwargs, 'data_loader', 'validation') # options : traininig, validation, testing
    data_loader = find_data_loader(type_data_loader, self._data)
    if getattr(self, 'name','') in ['MultiTaskRoadExtractor']:
        data_loader = find_data_loader(type_data_loader, self._orient_data)
        self._data.batch_size=1

    nodata = kwarg_fill_none(kwargs, 'nodata', 0)

    index = kwarg_fill_none(kwargs, 'index', 0)

    imsize = kwarg_fill_none(kwargs, 'imsize', 5)

    top = kwargs.get('top', None)
    title_font_size=16
    if top is None:
        top = get_top_padding(
            title_font_size=title_font_size,
            nrows=nrows,
            imsize=imsize
            )

    statistics_type = kwarg_fill_none(kwargs, 'statistics_type', 'dataset') # Accepted Values `dataset`, `DRA`
    stretch_type = kwargs.get('stretch_type', 'minmax') # Accepted Values `minmax`, `percentclip`

    # get batches
    x_batch, y_batch = get_nbatches(data_loader, math.ceil(nrows/self._data.batch_size))
    symbology_x_batch = x_batch = torch.cat(x_batch)
    if getattr(self, 'name','') in ['MultiTaskRoadExtractor']:
        y_batch= torch.stack([item for sublist in y_batch for item in sublist[0]]).type(torch.long).unsqueeze(1)
    else:
        y_batch = torch.cat(y_batch)

    symbology_bands = [0, 1, 2]
    if self._is_multispectral:
        # Get RGB Bands for plotting
        rgb_bands = kwarg_fill_none(kwargs, 'rgb_bands', self._data._symbology_rgb_bands)

        # Get Symbology bands
        symbology_bands = get_symbology_bands(rgb_bands, self._data._extract_bands, self._data._bands)

    # Get Predictions
    activation_store = []
    for i in range(0, x_batch.shape[0], self._data.batch_size):
        activations = predict_batch(self, x_batch[i:i+self._data.batch_size])
        activation_store.append(activations)

    # Analyze Pred
    if getattr(self, 'name','') in ['MultiTaskRoadExtractor']:
        activation_store = [x[0] for x in activation_store]
    predictions = analyze_pred_pixel_classification(self, activation_store)

    # Denormalize X
    x_batch = denorm_x(x_batch, self)

    # Extract RGB Bands for plotting
    symbology_x_batch = x_batch[:, symbology_bands]
    if stretch_type is not None:
        symbology_x_batch = image_batch_stretcher(symbology_x_batch, stretch_type, statistics_type)

    # Apply Image Strecthing
    if statistics_type == 'DRA':
        symbology_x_batch = dynamic_range_adjustment(symbology_x_batch)

    symbology_x_batch = image_tensor_checks_plotting(symbology_x_batch)

    # Get color Array
    color_array = self._data._multispectral_color_array
    color_array[1:, 3] = alpha

    # Size for plotting
    nrows = min(nrows, symbology_x_batch.shape[0])
    fig, axs = plt.subplots(nrows=nrows, ncols=ncols, figsize=(ncols*imsize, nrows*imsize))
    plt.subplots_adjust(top=top)
    fig.suptitle('Ground Truth / Predictions', fontsize=title_font_size)
    for r in range(nrows):
        if nrows==1:
            axi = axs
        else:
            axi  = axs[r]
        if r < symbology_x_batch.shape[0]:
            axi[0].imshow(symbology_x_batch[r].cpu().numpy())
            y_rgb = color_array[y_batch[r][0]]
            axi[0].imshow(y_rgb, alpha=alpha)
            axi[1].imshow(symbology_x_batch[r].cpu().numpy())
            p_rgb = color_array[predictions[r]]
            axi[1].imshow(p_rgb, alpha=alpha)
        axi[0].axis('off')
        axi[1].axis('off')
    #
    if return_fig:
        return fig, axs
